""" 
Element Line Plotter
--------------------

This module contains functions to plot element lines with pyqtgraph.
"""
import numpy as np
import tfs
import pyqtgraph as pg
import logging
from omc3.optics_measurements.constants import S
from qtpy.QtCore import Qt

LOGGER = logging.getLogger(__name__)
LENGTH: str = "LENGTH"

def plot_element_lines(plot: pg.PlotWidget, data_frame: tfs.TfsDataFrame, ranges: list[tuple[str, str]], start_zero: bool): 
    """ 
    Plot vertical lines on the plot for elements in the data_frame in the given ranges.

    Args:
        plot (pg.PlotWidget): The plot to plot the lines into.
        data_frame (tfs.TfsDataFrame): The data_frame to plot the lines from.
        ranges (list[tuple[str, str]]): A list of tuples of the form (start_element, end_element).
        start_zero (bool): Whether to start the plot from zero or not.  
    """
    if start_zero and not all(ranges[0][0] == r[0] for r in ranges):
        LOGGER.warning("Not all ranges start at the same element. Using only the first!")
        ranges = [ranges[0]]

    plotItem = plot.plotItem

    # Find start and end elements
    start = min(ranges, key=lambda r: data_frame.loc[r[0], S])[0]
    end = max(ranges, key=lambda r: data_frame.loc[r[1], S])[1]

    # Select element range and do some wrapping gymnastics if needed
    if data_frame.loc[start, S] <= data_frame.loc[end, S]:
        s_elements = data_frame.loc[start:end, S]
    else:
        s_elements = data_frame.loc[start:, S] + data_frame.loc[:end, S]
    
    if start_zero:
        s_elements = s_elements - s_elements.loc[start]
        s_elements = s_elements + np.where(s_elements < 0, data_frame.headers[LENGTH], 0)

    # Plot the lines - this takes a while, not sure how to improve
    plotItem.disableAutoRange()  # speeds it up a bit
    pen = pg.mkPen(color="grey", width=1, style=Qt.PenStyle.DotLine)
    for element, x in s_elements.items():
        plotItem.addLine(x=x, z=-10, pen=pen, label=element, labelOpts={"angle": 90})



""" 
UI: Colors
----------

This module contains the color definitions for the application.
All colors should be defined here for consistency and the elements
will then refer to the main color-constants, e.g. `colors.TEXT_DARK` 
(not to the colors themselves).
This allows for a unified look and feel.
"""
import logging

# Predefined Colors ---
BLACK: str = "#000000"
BLACK_87: str = "#212121"
BLACK_54: str = "#757575"
BLACK_38: str = "#9e9e9e"
BLACK_26: str = "#bdbdbd"
BLACK_12: str = "#e1e1e1"
WHITE: str = "#FFFFFF"

GREEN: str = "#00FF00"
GREEN_DARK: str = "#28642A"
GREEN_LIGHT: str = "#4CAF50"
GREEN_DARK_GREY: str = "#8EA18F"
GREEN_LIGHT_GREY: str = "#C8E6C9"

RED: str = "#FF0000"
RED_DARK: str = "#B71C1C"
RED_LIGHT: str = "#F44336"
RED_GREY: str = "#CFADAD"

BLUE: str = "#0000FF"
BLUE_DARK: str = "#0D47A1"
BLUE_LIGHT: str = "#2196F3"
BLUE_GREY: str = "#9FA8DA"

ORANGE_DARK: str = "#E65100"
ORANGE_LIGHT: str = "#FF9800"
ORANGE_GREY: str = "#F9E0B3"

PURPLE_DARK: str = "#4A148C"
PURPLE_LIGHT: str = "#9C27B0"
PURPLE_GREY: str = "#D1C4E9"

CORAL: str = "#E91E63"

# Machine Colors ---
# LHC Colors
BEAM1: str = BLUE
BEAM2: str = RED

# PSB Ring Colors
RING1: str = GREEN_LIGHT
RING2: str = ORANGE_LIGHT
RING3: str = PURPLE_DARK
RING4: str = CORAL

# Light Background, dark text ---
TEXT_DARK: str = BLACK_87
SECONDARY_TEXT_DARK: str = BLACK_54
GREYED_OUT_TEXT_DARK: str = BLACK_26

# Dark Background, light text ---
TEXT_LIGHT: str = WHITE
SECONDARY_TEXT_LIGHT: str = BLACK_12

# Tooltips ---
TOOLTIP_TEXT: str = BLACK_87
TOOLTIP_BACKGROUND: str = BLACK_12
TOOLTIP_BORDER: str = BLACK_38

# Logging ---
LOGGING: dict[int, str] = {
    logging.DEBUG: BLACK_26,        # default: black 
    logging.INFO: BLACK_87,         # default: green
    # logging.WARNING: ORANGE_LIGHT,  # default: orange
    # logging.ERROR: RED_DARK,        # default: red
    # logging.CRITICAL: CORAL,   
}


from pathlib import Path

this_dir = Path(__file__).parent

# Welcome ---
print("Welcome to omc3_gui!\n")

# Scripts ---
scripts = this_dir.glob("[a-zA-Z]*.py")
print("Available entrypoints:\n")
for script in scripts:
    print(f"omc3_gui.{script.stem}")
print()



""" 
DataClass UI: Tools 
-------------------

Additional tools that can be used with dataclasses.
"""
from __future__ import annotations

from dataclasses import Field, fields
import inspect
import json
import logging
from pathlib import Path
import re


LOGGER = logging.getLogger(__name__)


def get_field_inline_comments(dclass: type) -> dict[str, str]:
    """
    Returns a dictionary mapping field names to their associated inline code-comments.
    Has been replaced by the use of the metadata, but I like the function,
    so I leave it here. (jdilly 2023)

    Parameters:
        dclass (type): The data class to extract field comments from.

    Returns:
        Dict[str, str]: A dictionary mapping field names to their associated comments.
    """
    matcher = re.compile(r"^(?P<field>[a-zA-Z_]+)\s*:\s*[^#]+#\s*(?P<comment>.*)\s*$")
    source = inspect.getsource(dclass)

    found_fields = {}
    for line in source.splitlines()[2:]:  # first two is @dataclass and name
        line = line.strip()
        if line.startswith('def '):
            break

        match = matcher.match(line)
        if match:
            found_fields[match.group('field')] = match.group('comment')

    return found_fields


# JSON -------------------------------------------------------------------------

# Load ---

def load_dataclass_from_json(
    dclass: type, 
    json_file: str | Path
):
    """
    Load a dataclass from a JSON file.
    Selects only the fields that are in the dataclass. Ignores fields that start with an underscore.

    Parameters:
        dclass (type): The data class to load from the JSON file.
        json_file (str): The path to the JSON file to load the data class from.

    Returns:
        object: An instance of the data class loaded from the JSON file.
    """
    return dclass(**_load_json_file(json_file, dclass))


def update_dataclass_from_json(
    data: object, 
    json_file: str | Path
):
    """
    Update a dataclass with data from a JSON file.
    Selects only the fields that are in the dataclass and json file. 
    Ignores fields that start with an underscore.

    Parameters:
        dclass (object): The data class instance to update from to the JSON file.
        json_file (str): The path to the JSON file to load the data from.

    Returns:
        object: An instance of the data class loaded from the JSON file.
    """
    json_data = _load_json_file(json_file, data)
    for key, value in json_data.items(): 
        setattr(data, key, value)
    return data


def _load_json_file(json_file: str | Path, dclass: object | type | None = None) -> dict:
    """ Load a JSON file and return a dictionary of the data. 
    The data is converted to a path if the field type requires it and filtered if the field name 
    is either not in the dataclass or starts with an underscore. """
    data: dict = json.loads(json_file.read_text())
    if dclass is None:
        return data
    return {f.name: _maybe_path(data.get(f.name, None), f) for f in fields(dclass) if f.name[0] != "_"}


def _maybe_path(value: str | Path, field: Field) -> Path:
    """ Convert value to a path, if the field type requires it. """

    if isinstance(field.type, str):
        if "Path" in field.type and value is not None:
            return Path(value)
        return value
    
    if issubclass(field.type, Path) and value is not None:
        return Path(value)
    return value


# Save ----

def save_dataclass_to_json(
    data: object, 
    json_file: str | Path
):
    """
    Save a dataclass to a JSON file.
    Ignores fields that start with an underscore.

    Parameters:
        dclass (object): The data class instance to save to the JSON file.
        json_file (str): The path to the JSON file to save the data class to.
    """
    data = {f.name: getattr(data, f.name) for f in fields(data) if f.name[0] != "_"}
    for key, value in data.items():
        if isinstance(value, Path):
            data[key] = str(value)
    json_file.write_text(f"{json.dumps(data, indent=2)}\n")




""" 
DataClass UI: View 
------------------

The frontend components that make up the dataclass UI,
in particular the dialog windows that can be used to edit dataclasses.
"""
from __future__ import annotations

import logging
from typing import Protocol

from qtpy import QtWidgets

from omc3_gui.ui_components.dataclass_ui import controller

LOGGER = logging.getLogger(__name__)
class DataClassInterface(Protocol):
    """ Protocol for the DataClassDialog. 
    Observed by :class:`omc3_gui.utils.dataclass_ui.controller.DataClassUI` and 
    :class:`omc3_gui.utils.dataclass_ui.controller.DataClassTabbedUI`, 
    both of which can hence be used with the dialog.
    """
    layout: QtWidgets.QLayout
    model: object

    def update_ui(self): ...
    def update_model(self): ...
    def validate(self, only_modified: bool = False): ...
    def reset_labels(self): ...


class DataClassDialog(QtWidgets.QDialog):
    """ Simple dialog window to display the DataClassUI layout. 
    Adds some convenience functionality like "Ok" and "Cancel" buttons
    and automatic data-validation on close.
    """
    WINDOW_TITLE = "Edit DataClass"
    DEFAULT_SIZE = (800, 600)  # width, height, use -1 for auto
    
    def __init__(self, dataclass_ui: DataClassInterface, parent = None):
        super().__init__(parent)
        self._button_box: QtWidgets.QDialogButtonBox = None
        
        self._dataclass_ui: DataClassInterface = dataclass_ui
        self._build_gui()
        self._connect_signals()
        self._set_size(width=self.DEFAULT_SIZE[0], height=self.DEFAULT_SIZE[1])
        self.update_ui()
        self.validate_only_modified: bool = True


    def _set_size(self, width: int = -1, height: int = -1):
        # Set position to the center of the parent (does not work in WSL for me, jdilly 2023)
        # parent = self.parent()
        # if parent is not None:
        #     parent_geo = parent.geometry()
        #     parent_pos = parent.mapToGlobal(parent.pos())  # multiscreen support
        #     if width >= 0:
        #         x = parent_pos.x() + parent_geo.width() / 2
        #     else:
        #         x = parent_pos.x() + (parent_geo.width() - width) / 2

        #     if height >=0 :
        #         y = parent_pos.y() + parent_geo.height() / 2
        #     else:
        #         y = parent_pos.y() + (parent_geo.height() - height) / 2
        #     self.move(x, y)
        
        # Set size
        self.resize(width, height)

    def _build_gui(self):
        self.setWindowTitle(self.WINDOW_TITLE)
        layout = QtWidgets.QVBoxLayout()
        layout.addLayout(self._dataclass_ui.layout)

        QBtn = QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel
        self._button_box = QtWidgets.QDialogButtonBox(QBtn)
        layout.addWidget(self._button_box)

        self.setLayout(layout)
    
    def _connect_signals(self):
        self._button_box.accepted.connect(self.accept)
        self._button_box.rejected.connect(self.reject)

    def update_ui(self, new_model: object = None):
        if new_model is not None:
            self._dataclass_ui.model = new_model 
        self._dataclass_ui.update_ui()  # triggers changes, so the labels appear in "changed" state
        self._dataclass_ui.reset_labels()  # so we reset them

    def accept(self):
        try:
            self._dataclass_ui.validate(only_modified=self.validate_only_modified)
        except ValueError as e:
            QtWidgets.QMessageBox.critical(self, "Error", str(e))
            return

        self._dataclass_ui.update_model()
        super().accept() 


class SettingsDialog(DataClassDialog):
    """ Slight modification of the DataClassDialog to be used for Tabbed-Settings. """

    WINDOW_TITLE = "Settings"
    DEFAULT_SIZE = (800, -1)
    
    def __init__(self, settings: object, parent=None):
        dataclass_tabbed_ui = controller.DataClassTabbedUI(dclass=settings)       
        super().__init__(dataclass_ui=dataclass_tabbed_ui, parent=parent)

    @property
    def settings(self) -> object:
        return self._dataclass_ui.model


# Type-to-Widget Helpers ----------------------------------------------------------------

class QFullIntSpinBox(QtWidgets.QSpinBox):
    """ Like a QSpinBox, but overwriting default range(0,100) with maximum integer range. """

    def __init__(self, parent: QtWidgets.QWidget | None = None) -> None:
        super().__init__(parent)
        self.setRange(-2**31, 2**31 - 1)  # range of signed 32-bit integers 





""" 
Main View
---------

This is the main view for the Segment-by-Segment application.
"""
# from omc3_gui.segment_by_segment.segment_by_segment_ui import Ui_main_window
from __future__ import annotations

import logging
from collections.abc import Sequence
from dataclasses import fields
from functools import partial

from omc3.definitions.optics import (
    ALPHA_COLUMN,
    BETA_COLUMN,
    DISPERSION_COLUMN,
    PHASE_COLUMN,
)
from qtpy import QtGui, QtWidgets
from qtpy.QtCore import QItemSelectionModel, QModelIndex, Qt, Signal, Slot

from omc3_gui.plotting.classes import DualPlotWidget
from omc3_gui.segment_by_segment.help_view import show_help_dialog
from omc3_gui.segment_by_segment.main_model import (
    MeasurementListModel,
    SegmentTableModel,
)
from omc3_gui.segment_by_segment.measurement_model import OpticsMeasurement
from omc3_gui.segment_by_segment.plotting import DualPlotDefinition
from omc3_gui.segment_by_segment.segment_model import SegmentItemModel
from omc3_gui.ui_components import colors
from omc3_gui.ui_components.base_classes_cvm import View
from omc3_gui.ui_components.styles import MONOSPACED_TOOLTIP
from omc3_gui.ui_components.widgets import (
    ChangeButton,
    DefaultButton,
    OpenButton,
    RemoveButton,
    RunButton,
)
from omc3_gui.utils.counter import HorizontalGridLayoutFiller
from omc3_gui.utils.iteration_classes import IterClass

ItemDataRole = Qt.ItemDataRole
LOGGER = logging.getLogger(__name__)


class Tabs(IterClass):
    """ Define the Tabs and the things to plot in them. """
    PHASE: DualPlotDefinition = DualPlotDefinition.generate_xy("Phase", "phase", PHASE_COLUMN)
    BETA: DualPlotDefinition = DualPlotDefinition.generate_xy("Beta", "beta_phase", BETA_COLUMN) 
    ALPHA: DualPlotDefinition = DualPlotDefinition.generate_xy("Alpha", "alpha_phase", ALPHA_COLUMN) 
    DISPERSION: DualPlotDefinition = DualPlotDefinition.generate_xy("Dispersion", "dispersion", DISPERSION_COLUMN)
    F1001AP: DualPlotDefinition = DualPlotDefinition.generate_amplitude_phase("f1001")
    F1001RI: DualPlotDefinition = DualPlotDefinition.generate_real_imag("f1001")
    F1010AP: DualPlotDefinition = DualPlotDefinition.generate_amplitude_phase("f1010")
    F1010RI: DualPlotDefinition = DualPlotDefinition.generate_real_imag("f1010")
    

class SbSWindow(View):
    WINDOW_TITLE = "OMC Segment-by-Segment"

    # QtSignals need to be defined as class-attributes
    sig_list_measurements_double_clicked = Signal(OpticsMeasurement)
    sig_list_measurements_selected = Signal(tuple)  # Tuple[OpticsMeasurement]
    sig_table_segments_selected = Signal(tuple)
    sig_thread_spinner_double_clicked = Signal()
    sig_tab_changed = Signal()

    # Menu Signals ---
    sig_menu_settings = Signal()
    sig_menu_clear_all = Signal()
    
    def __init__(self, parent=None):
        
        super().__init__(parent)
        self.setWindowTitle(self.WINDOW_TITLE)
        
        # List of UI elements accessible as instance-attributes:
        # Widgets ---
        self._cental: QtWidgets.QSplitter = None
        self._tabs_widget: QtWidgets.QTabWidget = None
        self._list_view_measurements: QtWidgets.QListView = None
        self._table_segments: QtWidgets.QTableView = None

        # Buttons ---
        self.button_load_measurement: QtWidgets.QPushButton = None
        self.button_remove_measurement: QtWidgets.QPushButton = None
        self.button_edit_measurement: QtWidgets.QPushButton = None
        self.button_copy_measurement: QtWidgets.QPushButton = None
        self.button_edit_corrections: QtWidgets.QPushButton = None
        self.button_run_matcher: QtWidgets.QPushButton = None

        self.button_run_segment: QtWidgets.QPushButton = None
        self.button_remove_segment: QtWidgets.QPushButton = None
        self.button_copy_segment: QtWidgets.QPushButton = None
        self.button_new_segment: QtWidgets.QPushButton = None
        self.button_default_segments: QtWidgets.QPushButton = None
        self.button_save_segments: QtWidgets.QPushButton = None
        self.button_load_segments: QtWidgets.QPushButton = None

        # Build GUI and connect Signals ---     
        self._add_menus()
        self._build_gui()
        self._connect_signals()

    def _connect_signals(self):
        """ Connect signals with slots. 
            Here only the basic signals are connected, the ones requireing more complex logic
            or calling different views are connected in the controller. 
        """
        # Optics Measurements ---
        self._list_view_measurements.doubleClicked.connect(self._handle_list_measurements_double_clicked)
        self._list_view_measurements.selectionModel().selectionChanged.connect(self._handle_list_measurements_selected)
        self._tabs_widget.currentChanged.connect(self._handle_tab_changed)

    # Slots --------------------------------------------------------------------
    @Slot(QModelIndex)
    def _handle_list_measurements_double_clicked(self, idx):
        LOGGER.debug(f"Entry in Optics List double-clicked: {idx.data(role=Qt.DisplayRole)}")
        self.sig_list_measurements_double_clicked.emit(idx.data(role=Qt.UserRole))

    @Slot()
    def _handle_list_measurements_selected(self):        
        LOGGER.debug("Optics List selection changed.")
        selected_measurements = self.get_selected_measurements()
        self.sig_list_measurements_selected.emit(selected_measurements)

    @Slot()
    def _handle_table_segments_selected(self):
        LOGGER.debug("Segment Table selection changed.")
        selected_segments = self.get_selected_segments()
        self.sig_table_segments_selected.emit(selected_segments)
    
    @Slot()
    def _handle_tab_changed(self):
        LOGGER.debug("Tab changed.")
        self.sig_tab_changed.emit()

    # Menu ---------------------------------------------------------------------
    def _add_menus(self):
        # File ---
        file_menu: QtWidgets.QMenu = self.get_action_by_title("File")  # defined in View-class
        file_menu.setTitle("SbS-GUI")

        # Settings ---
        menu_settings = QtWidgets.QAction("Settings", self)
        menu_settings.setIcon(
            QtWidgets.QApplication.style().standardIcon(QtWidgets.QStyle.StandardPixmap.SP_ComputerIcon)
        )
        menu_settings.triggered.connect(self.sig_menu_settings.emit)

        # insert before the last entry (which is "Exit")
        file_menu.insertAction(file_menu.actions()[-1], menu_settings)

        # Help --- 
        help_menu: QtWidgets.QMenu = self.get_action_by_title("Help")  # defined in View-class

        # Clear All -
        menu_clear_all = QtWidgets.QAction("Reload Data", self)
        menu_clear_all.setIcon(
            QtWidgets.QApplication.style().standardIcon(QtWidgets.QStyle.StandardPixmap.SP_DialogResetButton)
        )
        menu_clear_all.triggered.connect(self.sig_menu_clear_all.emit)
        help_menu.insertAction(help_menu.actions()[-1], menu_clear_all)

        menu_show_help = QtWidgets.QAction("Show Help", self)
        menu_show_help.setIcon(
            QtWidgets.QApplication.style().standardIcon(QtWidgets.QStyle.StandardPixmap.SP_MessageBoxQuestion)
        )
        menu_show_help.triggered.connect(show_help_dialog)  # more of a controller thing, but OK for this one
        help_menu.insertAction(help_menu.actions()[-1], menu_show_help)

    def add_settings_to_menu(self, menu: str, settings: object, names: Sequence[str] | None = None, hook: callable = None):
        """ Add quick-access checkboxes to the menu which are connected to the respective attributes in settings.
        
        Args:
            menu (str): Main menu name to add the settings to.
            settings (object): Settings to connect with the menu. Assumes dataclasses.
            names (Sequence[str] | None): Which fields to connect. All fields need to be boolean.
            hook (callable | None): Function to call after the settings have been updated.

        """
        qmenu: QtWidgets.QMenu = self.get_action_by_title(menu)
        
        def update_settings(value: bool, name: str):
            setattr(settings, name, value)
            if hook is not None:
                hook()

        qmenu.addSeparator()
        for field in fields(settings): 
            if names is not None and field.name not in names:
                continue

            if field.name.startswith("_"):
                continue
            
            value = getattr(settings, field.name)
            if not isinstance(value, bool):
                continue

            label = field.metadata.get("label", field.name)
            entry = QtWidgets.QAction(label, self)
            entry.setCheckable(True)
            entry.setChecked(value)
            entry.toggled.connect(partial(update_settings, name=field.name))
            qmenu.addAction(entry)
        qmenu.addSeparator()

    def update_menu_settings(self, menu: str, settings: object, names: Sequence[str] | None = None):
        """ Update the menu settings. 
        See :func:`add_settings_to_menu`.
        
        Args:
            menu (str): Main menu name where the settings are located.
            settings (object): Settings to connected with the menu. Assumes dataclasses.
            names (Sequence[str] | None): Which fields to connect. All fields need to be boolean.
        """
        qmenu: QtWidgets.QMenu = self.get_action_by_title(menu)

        for field in fields(settings): 
            if names is not None and field.name not in names:
                continue

            if field.name.startswith("_"):
                continue

            label = field.metadata.get("label", field.name)
            entry: QtWidgets.QAction = self.get_action_by_title(label, parent=qmenu)
            if entry is None:
                continue

            entry.setChecked(getattr(settings, field.name))

    # Build Main UI-------------------------------------------------------------
    def _build_gui(self):
        self._central = QtWidgets.QSplitter(Qt.Horizontal)

        def build_navigation_widget():  # --- Left Hand Side 
            navigation_widget = QtWidgets.QSplitter(Qt.Vertical)

            def build_navigation_top():
                nav_top = QtWidgets.QWidget()

                layout = QtWidgets.QVBoxLayout()
                nav_top.setLayout(layout)
                layout.addWidget(QtWidgets.QLabel("Loaded Optics:"))

                self._list_view_measurements = MeasurementListView()
                layout.addWidget(self._list_view_measurements)

                def build_measurement_buttons():
                    grid_buttons = QtWidgets.QGridLayout()
                    grid_buttons_filler = HorizontalGridLayoutFiller(layout=grid_buttons, cols=3)

                    load = OpenButton("Load")
                    load.setToolTip("Load a measurement, i.e. omc3-optics output folder.")
                    grid_buttons_filler.add(load)
                    self.button_load_measurement = load
                    
                    copy = DefaultButton("Copy")
                    copy.setToolTip("Create a virtual copy of the measurement, with a different output dir.")
                    grid_buttons_filler.add(copy)
                    self.button_copy_measurement = copy

                    remove = RemoveButton()
                    remove.setToolTip("Remove the currently selected measurement(s).")
                    grid_buttons_filler.add(remove)
                    self.button_remove_measurement = remove

                    matcher = RunButton("Run Matcher")
                    matcher.setToolTip("Run the Segment-by-Segment Matcher.")
                    grid_buttons_filler.add(matcher)
                    self.button_run_matcher = matcher
                    
                    edit = DefaultButton("Edit")
                    edit.setToolTip("Edit the settings of the currently selected measurement.")
                    grid_buttons_filler.add(edit)
                    self.button_edit_measurement = edit

                    edit_corrections = ChangeButton("Corrections")
                    edit_corrections.setToolTip("Edit the corrections file of the currently selected measurement.")
                    grid_buttons_filler.add(edit_corrections)
                    self.button_edit_corrections = edit_corrections
                
                    return grid_buttons

                layout.addLayout(build_measurement_buttons())
                return nav_top
            navigation_widget.addWidget(build_navigation_top())

            def build_navigation_bottom():
                nav_bottom = QtWidgets.QWidget()

                layout = QtWidgets.QVBoxLayout()
                nav_bottom.setLayout(layout)

                layout.addWidget(QtWidgets.QLabel("Segments:"))

                self._table_segments = SegmentTableView()
                layout.addWidget(self._table_segments)

                def build_segment_buttons():
                    grid_buttons = QtWidgets.QGridLayout()
                    grid_buttons_filler = HorizontalGridLayoutFiller(layout=grid_buttons, cols=3)

                    run = RunButton("Run Segment(s)")
                    run.setToolTip("Run the currently selected segment(s).")
                    grid_buttons_filler.add(run, col_span=3)
                    self.button_run_segment = run

                    new = OpenButton("New")
                    new.setToolTip("Add a new segment.")
                    grid_buttons_filler.add(new)
                    self.button_new_segment = new
                    
                    default = ChangeButton("Add Defaults")
                    default.setToolTip(
                        "Add default segments for the currently selected measurements (if not already present)."
                    )
                    grid_buttons_filler.add(default)
                    self.button_default_segments = default

                    copy = DefaultButton("Copy")
                    copy.setToolTip("Create a copy of the currently selected segment(s).")
                    grid_buttons_filler.add(copy)
                    self.button_copy_segment = copy

                    remove = RemoveButton("Remove")
                    remove.setToolTip("Remove the currently selected segment(s) from the list (does not delete files).")
                    grid_buttons_filler.add(remove)
                    self.button_remove_segment = remove
                    
                    save = DefaultButton("Save")
                    save.setToolTip("Save current segments definitions to a json file.")
                    grid_buttons_filler.add(save)
                    self.button_save_segments = save
                    
                    load = DefaultButton("Load")
                    load.setToolTip("Load segments definitions from an existing SbS-Folder or a json file.")
                    grid_buttons_filler.add(load)
                    self.button_load_segments = load
                
                    return grid_buttons
                layout.addLayout(build_segment_buttons())
                return nav_bottom

            navigation_widget.addWidget(build_navigation_bottom())
            return navigation_widget
        self._central.addWidget(build_navigation_widget())

        def build_tabs_widget():  # --- Right Hand Side
            self._tabs_widget = QtWidgets.QTabWidget()
            for tab in Tabs.values():
                tab: DualPlotDefinition
                self._tabs_widget.addTab(DualPlotWidget(), tab.name)
            return self._tabs_widget

        self._central.addWidget(build_tabs_widget())

        # Set up main widget layout ----
        self._central.setSizes([300, 1000])
        self._central.setStretchFactor(1, 3)
        
        self.setCentralWidget(self._central)

    # Interactors --------------------------------------------------------------
    def get_current_tab(self) -> tuple[DualPlotDefinition, DualPlotWidget]:
        widget = self._tabs_widget.currentWidget()
        index = self._tabs_widget.currentIndex()
        return list(Tabs.values())[index], widget

    # Getters and Setters
    def set_measurements_list(self, measurement_model: MeasurementListModel):
        self._list_view_measurements.setModel(measurement_model)

    def get_measurement_list(self) -> MeasurementListModel:
        return self._list_view_measurements.model()

    def get_selected_measurements(self) -> tuple[OpticsMeasurement]:
        """ Get the currently selected measurements from the GUI. 
        Hint: Use the Qt.UserRole to retrieve the actual OpticsMeasurement.
        """
        selected = self._list_view_measurements.selectedIndexes()
        return tuple(s.data(role=ItemDataRole.UserRole) for s in selected)
    
    def get_all_measurements(self) -> tuple[OpticsMeasurement]:
        """ Get the currently selected measurements from the GUI. 
        Hint: Use the Qt.UserRole to retrieve the actual OpticsMeasurement.
        """
        return self._list_view_measurements.model().items

    def set_selected_measurements(self, indices: Sequence[QModelIndex] = ()):
        self._list_view_measurements.selectionModel().clear()
        for idx in indices:
            self._list_view_measurements.selectionModel().select(idx, QItemSelectionModel.Select)

    def set_segments_table(self, segment_model: SegmentTableModel):
        self._table_segments.setModel(segment_model)
        self._table_segments.selectionModel().selectionChanged.connect(self._handle_table_segments_selected)

    def get_segments_table(self) -> SegmentTableModel:
        return self._table_segments.model()

    def get_selected_segments(self) -> tuple[SegmentItemModel]:
        """ Get the currently selected segments from the GUI. 
        Hint: Use the Qt.UserRole to retrieve the actual SegmentItemModel.
        """
        selected: list[QModelIndex] = self._table_segments.selectedIndexes()
        return tuple(s.data(role=ItemDataRole.UserRole) for s in selected if s.column() == 0)  # need only one per row
    

class MeasurementListView(QtWidgets.QListView):
    """ Defines the view for the measurement list (on the top left). """
    
    def __init__(self):
        super().__init__()
        self.setModel(MeasurementListModel())
        self.setItemDelegate(ColoredItemDelegate())
        self.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        self.setStyleSheet(MONOSPACED_TOOLTIP)


class SegmentTableView(QtWidgets.QTableView):
    """ Defines the view for the segment table (on the bottom left). """

    def __init__(self):
        super().__init__()
        self.setModel(SegmentTableModel())

        header_hor = self.horizontalHeader()
        header_hor.setSectionResizeMode(QtWidgets.QHeaderView.ResizeToContents)       
        header_hor.setSectionResizeMode(0, QtWidgets.QHeaderView.Stretch)

        header_ver = self.verticalHeader()
        header_ver.setVisible(False)
        header_ver.setSectionResizeMode(QtWidgets.QHeaderView.ResizeToContents)

        self.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        self.setShowGrid(True)
        self.setStyleSheet(MONOSPACED_TOOLTIP)


class ColoredItemDelegate(QtWidgets.QStyledItemDelegate):
    """ Defines an ItemDelegate that uses a custom color for the text. """

    COLOR_MAP = {
        MeasurementListModel.ColorIDs.NONE: colors.TEXT_DARK,
        MeasurementListModel.ColorIDs.BEAM1: colors.BEAM1,
        MeasurementListModel.ColorIDs.BEAM2: colors.BEAM2,
        MeasurementListModel.ColorIDs.RING1: colors.RING1,
        MeasurementListModel.ColorIDs.RING2: colors.RING2,
        MeasurementListModel.ColorIDs.RING3: colors.RING3,
        MeasurementListModel.ColorIDs.RING4: colors.RING4,
    }
    def paint(self, painter, option, index):
        # Customize the text color
        color_id = index.data(Qt.TextColorRole)
        try:
            color = self.COLOR_MAP[color_id]
        except KeyError:
            color = self.COLOR_MAP[MeasurementListModel.ColorIDs.NONE]
        option.palette.setColor(QtGui.QPalette.Text, QtGui.QColor(color))
        
        super().paint(painter, option, index)





""" 
Main Controller
---------------

This is the main controller for the Segment-by-Segment application.
"""
from __future__ import annotations

import logging
import re
from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING

from omc3.sbs_propagation import segment_by_segment
from qtpy import QtWidgets
from qtpy.QtCore import Slot

from omc3_gui.plotting.classes import DualPlotWidget
from omc3_gui.segment_by_segment.defaults import (
    DEFAULT_SEGMENTS,
    get_default_correctors,
)
from omc3_gui.segment_by_segment.main_model import SegmentTableModel
from omc3_gui.segment_by_segment.main_view import SbSWindow
from omc3_gui.segment_by_segment.measurement_model import OpticsMeasurement
from omc3_gui.segment_by_segment.measurement_view import OpticsMeasurementDialog
from omc3_gui.segment_by_segment.plotting import plot_segment_data
from omc3_gui.segment_by_segment.segment_model import (
    SegmentDataModel,
    SegmentItemModel,
    compare_segments,
    get_segments_from_directory,
)
from omc3_gui.segment_by_segment.segment_view import SegmentDialog
from omc3_gui.segment_by_segment.settings import PlotSettings, Settings
from omc3_gui.ui_components.base_classes_cvm import Controller
from omc3_gui.ui_components.dataclass_ui import SettingsDialog
from omc3_gui.ui_components.file_dialogs import OpenAnyMultiDialog, OpenAnySingleDialog
from omc3_gui.ui_components.message_boxes import show_confirmation_dialog
from omc3_gui.ui_components.text_editor import TextEditorDialog
from omc3_gui.ui_components.threads import BackgroundThread

if TYPE_CHECKING:
    from collections.abc import Sequence

LOGGER = logging.getLogger(__name__)

class SbSController(Controller):
    
    settings: Settings
    _view: SbSWindow

    def __init__(self, measurements: list[Path | str] | None = None, settings: Settings | None = None):
        super().__init__(SbSWindow())
        self.settings: Settings = settings or Settings()
        self._last_selected_measurement_path: Path = self.settings.main.cwd
        self._running_tasks: list[BackgroundThread] = []

        self.connect_signals()
        self.set_measurement_interaction_buttons_enabled(False)
        self.set_all_segment_buttons_enabled(False)
        
        if measurements is not None:
            self.open_measurements_from_paths(measurements)        

    def connect_signals(self):
        """ Connect the signals from the GUI components (view) to the slots (controller). """
        view: SbSWindow = self._view  # for shorthand and type hinting

        # Tabs -----------------------------------------------------------------
        view.sig_tab_changed.connect(self.plot)

        # Menu Bar -------------------------------------------------------------
        view.sig_menu_settings.connect(self.show_settings)

        view.add_settings_to_menu(
            menu="View",
            settings=self.settings.plotting,
            hook=self.plot,
        )

        view.sig_menu_clear_all.connect(self.clear_all_data)

        # Measurements -------------------------------------------------------------
        view.button_load_measurement.clicked.connect(self.open_measurements)
        view.button_edit_measurement.clicked.connect(self.edit_measurement)
        view.button_remove_measurement.clicked.connect(self.remove_measurement)
        view.button_copy_measurement.clicked.connect(self.copy_measurement)

        view.button_run_matcher.clicked.connect(self.run_matcher)
        view.button_edit_corrections.clicked.connect(self.edit_corrections)

        view.sig_list_measurements_double_clicked.connect(self.edit_measurement)
        view.sig_list_measurements_selected.connect(self.measurement_selection_changed)

        # Segments -------------------------------------------------------------
        view.button_new_segment.clicked.connect(self.new_segment)
        view.button_copy_segment.clicked.connect(self.copy_segment)
        view.button_default_segments.clicked.connect(self.add_default_segments)
        view.button_remove_segment.clicked.connect(self.remove_segment)
        view.button_run_segment.clicked.connect(self.run_segments)
        view.button_save_segments.clicked.connect(self.save_segments)
        view.button_load_segments.clicked.connect(self.load_segments)

        view.sig_table_segments_selected.connect(self.segment_selection_changed)
        view.sig_thread_spinner_double_clicked.connect(self._show_running_tasks)

    # Tasks --------------------------------------------------------------------  
    @Slot()
    def _update_tasks_status(self):
        """ Update the status bar with the number of running tasks. """
        view: SbSWindow = self._view  
        status_bar: QtWidgets.QStatusBar = view.statusBar()  # seems to return it, if already exist, because spinner is there
        
        if self._running_tasks:
            # status_bar.show()  # looks nice, but moves the window around too much ... 
            status_bar.showMessage(f"{len(self._running_tasks)} Task(s) running ...")
            status_bar.setToolTip(
                f"{len(self._running_tasks)} Running Task(s):\n  - "
                + "\n  - ".join([task.message for task in self._running_tasks])
            )
            view.thread_spinner.start()
        else:
            status_bar.setToolTip(None)
            status_bar.clearMessage()
            view.thread_spinner.stop()
            # status_bar.hide()

    @Slot()
    def _add_running_task(self, task: BackgroundThread):
        """ Add a task to the list of running tasks. """
        
        # Automatically remove task when finished
        remove_task_fun = partial(self._remove_running_task, task=task)
        task.finished.connect(remove_task_fun)

        self._running_tasks.append(task)
        self._update_tasks_status()

    @Slot()
    def _remove_running_task(self, task: BackgroundThread):
        """ Remove a task from the list of running tasks. """
        self._running_tasks.remove(task)
        self._update_tasks_status()
        if not self._running_tasks:
            self.clear_all_data() # last task finished, update plots

    @Slot()
    def _show_running_tasks(self):
        """ Show (i.e. log) the list of running tasks. """
        LOGGER.info(f"Running tasks: {[task.message for task in self._running_tasks]}")
        
    # Measurements -------------------------------------------------------------
    def set_measurement_interaction_buttons_enabled(self, enabled: bool = True):
        """ Enable/disable the buttons that interact with measurements. 
        
        Args:
            enabled (bool): True to enable, False to disable.
        """
        view: SbSWindow = self._view  

        measurement_interaction_buttons = (
            view.button_remove_measurement,
            view.button_edit_measurement,
            view.button_run_matcher,
            view.button_edit_corrections,
            view.button_copy_measurement,
        )
        for button in measurement_interaction_buttons:
            button.setEnabled(enabled)

    def add_measurement(self, measurement: OpticsMeasurement):
        """ Add a measurement to the GUI. 
        
        Args:
            measurement (OpticsMeasurement): The measurement to add.
        """
        view: SbSWindow = self._view  # for type hinting
        view.get_measurement_list().add_item(measurement) 
    
    @Slot()
    def open_measurements(self):
        """ Open the file dialog for optics measurements. """
        view: SbSWindow = self._view  

        LOGGER.debug("Opening new optics measurement. Asking for paths.")
        filenames = OpenAnyMultiDialog(
            parent=view,
            caption="Select Optics Folders/SbS Folders/SbS json", 
            directory=self._last_selected_measurement_path,
        ).run_selection_dialog()

        self.open_measurements_from_paths(filenames, select=True)

    def open_measurements_from_paths(self, paths: Sequence[Path | str], select: bool = False):
        """ Open the given paths as measurements. """
        if not len(paths):
            LOGGER.debug("No measurement paths to load.")
            return

        view: SbSWindow = self._view
        loaded_measurements = view.get_measurement_list()
        measurement_indices = []

        for filename in paths:
            LOGGER.debug(f"Adding: {filename!s}")
            if filename.is_dir():
                optics_measurement = OpticsMeasurement.from_path(filename)
            elif filename.is_file() and filename.suffix == ".json":
                optics_measurement = OpticsMeasurement.from_json(filename)
            else:
                LOGGER.error(f"Invalid file: {filename}")
                continue

            try:
                optics_measurement.quick_check()
            except ValueError as e:
                LOGGER.warning(str(e))  # Maybe even popup?
            except NameError as e:
                LOGGER.error(f"{e!s} ({filename})")
                continue

            if self.settings.main.autoload_segments:
                self.load_segments_for_measurement(optics_measurement)

            if self.settings.main.autodefault_segments:
                self.add_default_segments(optics_measurement)

            try:
                loaded_measurements.add_item(optics_measurement)
            except ValueError as e:
                LOGGER.error(str(e))
            else:
                if select: 
                    measurement_indices.append(loaded_measurements.get_index(optics_measurement))

            self._last_selected_measurement_path = filename.parent

        view.set_selected_measurements(measurement_indices)
    
    @Slot()
    def edit_measurement(self, measurement: OpticsMeasurement | None = None):
        """ Open the edit dialog for a measurement.
        If no measurement is given, the currently selected measurement is used.

        Args:
            measurement (OpticsMeasurement | None, optional): The measurement to edit. Defaults to None.
        """
        if measurement is None:
            try:
                measurement = self.get_single_measurement()
            except ValueError as e:
                LOGGER.warning(str(e))
                return

        LOGGER.debug(f"Opening edit dialog for {measurement.display()}.")
        view: SbSWindow = self._view  
        dialog = OpticsMeasurementDialog(
            parent=view,
            optics_measurement=measurement,
        )
        if dialog.exec_() == dialog.Accepted:
            LOGGER.debug("Edit dialog closed. Updating measurement.")

        try:
            measurement.quick_check()
        except ValueError as e:
            LOGGER.warning(str(e))  # Maybe even popup?
        except NameError as e:
            LOGGER.error(str(e))
            return
        
        try:
            measurement.to_json()
        except IOError as e:
            LOGGER.warning(str(e))


    @Slot()
    def copy_measurement(self, measurement: OpticsMeasurement | None = None):
        """ Create a copy of the given measurement and add it to the GUI. 
        If no measurement is given, the currently selected measurement is copied.

        Args:
            measurement (OpticsMeasurement | None, optional): The measurement to copy. Defaults to None.
        """
        if measurement is None:
            try:
                measurement = self.get_single_measurement()
            except ValueError as e:
                LOGGER.warning(str(e))
                return

        if measurement.output_dir is None:  # should not be possible, but better to catch
            LOGGER.error("Cannot copy measurement without output directory.")
            return

        LOGGER.debug(f"Copying {measurement.display()}.")
        new_measurement = measurement.copy()
        
        # might already have a counter from previous copy
        name =  re.sub(r"_\d+$", "", measurement.output_dir.name)  

        for count in range(1, 1000):  # limit to avoid infinite loop
            new_measurement.output_dir = measurement.output_dir.with_name(f"{name}_{count:d}")
            try:
                self.add_measurement(new_measurement)
            except ValueError:
                continue
            break
        else:
            LOGGER.error(
                "Could not copy measurement. Counter limit exeeded. Not sure what went wrong."
            )
            return

    @Slot()
    def remove_measurement(self, measurements: Sequence[OpticsMeasurement] | None = None):
        """ Remove measurements from the GUI.
        If no measurements are given, the currently selected measurements are removed.        

        Args:
            measurements (Sequence[OpticsMeasurement] | None, optional): The measurements to remove. Defaults to None.
        """
        view: SbSWindow = self._view  

        if measurements is None:
            measurements = view.get_selected_measurements()
            if not len(measurements):
                LOGGER.warning("No measurement selected.")
                return
        
        view.get_measurement_list().remove_items(measurements)
        view.set_selected_measurements()

    @Slot(tuple)
    def measurement_selection_changed(self, measurements: Sequence[OpticsMeasurement]):
        """ Updates the GUI when the selected measurements change. 
        
        Args:
            measurements: Sequence[OpticsMeasurement]: The new selection of measurements.
        """
        LOGGER.debug(f"Selected {len(measurements)} measurements.")
        view: SbSWindow = self._view  

        if not len(measurements):
            self.set_measurement_interaction_buttons_enabled(False)
            view.set_segments_table(SegmentTableModel())
            self.segment_selection_changed()
            self.set_all_segment_buttons_enabled(False)
            return

        self.set_measurement_interaction_buttons_enabled(True)
        if len(measurements) > 1:
            view.button_edit_measurement.setEnabled(False)
            view.button_copy_measurement.setEnabled(False)

        self.set_all_segment_buttons_enabled(True)

        # Group the segments for the measurements into table-items when they have the same defintion ---
        segment_table_items: list[SegmentItemModel] = []

        for measurement in measurements:
            for segment in measurement.segments:
                for segment_item in segment_table_items:
                    if compare_segments(segment, segment_item):
                        segment_item.append_segment(segment)
                        break
                else:
                    segment_table_items.append(SegmentItemModel.from_segments([segment]))

        # Create the segment Table to show in the GUI ---
        segment_table = SegmentTableModel()
        try:
            segment_table.add_items(segment_table_items)
        except ValueError as e:
            LOGGER.debug(str(e))
            
        view.set_segments_table(segment_table)
        self.segment_selection_changed()

    def get_single_measurement(self) -> OpticsMeasurement:
        """ Get a single selected measurement from the GUI. 
        Raises ValueError if no measurement is selected or multiple measurements are selected. 
        """
        view: SbSWindow = self._view  
        measurements = view.get_selected_measurements()

        if len(measurements) == 0:
            raise ValueError("Please select at least one measurement.")

        if len(measurements) > 1:
            raise ValueError("Please select only one measurement.")
        
        return measurements[0]

    @ Slot()
    def run_matcher(self) -> None:
        """ Run the matcher. """
        view: SbSWindow = self._view
        view.showErrorDialog("Error: Not Implemented", "The Segment-by-Segment Matcher is not implemented yet.")
        # TODO!
    
    @ Slot()
    def edit_corrections(self) -> None:
        """ Edit the corrections file. 
        
        The following logic is applied to the selected measurements:
        
        - Check if a measurement is selected:
        - If not show error.

        - Check if multiple measurements are selected, if so: 
            a) they all have the same correction file: open TextEditor, 
            c) they have different correction files: show error
            b) some have the same correction file, all others have none: ask if the others should also get this one
            d) they have no correction file: ask for path and then open TextEditor with that file
        - If only a single measurement is s
        """
        view: SbSWindow = self._view

        selected_measurements: tuple[OpticsMeasurement] = view.get_selected_measurements()
        if not selected_measurements:
            LOGGER.error("Please select at least one measurement.")
            return
        
        correction_files = {measurement.corrections for measurement in selected_measurements if measurement.corrections}
        if len(correction_files) > 1:
            view.showErrorDialog(
                title="Error: Multiple correction files", 
                message="Please select only measurements using the same correction file."
            )
            return
        
        # Only one or none correction file within the selection measurements from here ---
        if len(correction_files) == 0:  # If there is none, ask user to provide one
            LOGGER.debug("No correction file selected. Asking.")
            directory = self.settings.main.cwd
            if len(selected_measurements) == 1:
                directory = selected_measurements[0].measurement_dir

            dialog = OpenAnySingleDialog(
                caption="Select a new or existing correction file for the selected measurement(s).",
                existing=False,  # can be a new file
                parent=view,
                directory=directory
            )
            correction_file = dialog.run_selection_dialog()
            if correction_file is None:
                LOGGER.error("No correction file to edit selected.")
                return

        else:  # There is only one. Maybe use it for all selected measurements
            correction_file = correction_files.pop()

            has_no_corrections = [m for m in selected_measurements if not m.corrections]
            if has_no_corrections:
                measurements_string = '\n'.join([m.display() for m in has_no_corrections])
                use_for_all = show_confirmation_dialog(
                    question=(
                        f"The measurements\n\n{measurements_string}\n\n"
                        "have no correction file assigned.\n"
                        f"Do you want to use\n\n{correction_file}\n\n"
                        "also for these measurements?"
                    ),
                    title="There are unset correction files",
                )
                if not use_for_all:
                    LOGGER.error(
                        "User does not want to use the same correction file for all selected measurements."
                        "Select a correction file for each measurement manually."
                    )
                    return
                
        # Use the correction file for all selected measurements ---
        # We could simply assign the correction file to all (as the other have the same file),
        # but this way we can log the changes (if even any).
        for measurement in selected_measurements:
            if not measurement.corrections:
                LOGGER.debug(f"Setting {correction_file} for {measurement.display()}")
                measurement.corrections = correction_file

        # Open the TextEditor for the selected correction file ---
        LOGGER.debug(f"Opening TextEditor for correction file: {correction_file}.")
        edit_dialog = TextEditorDialog(correction_file)

        if not correction_file.exists() and self.settings.main.suggest_correctors:
            text = get_default_correctors(selected_measurements[0])  # bit hacky but ok for now?
            edit_dialog.text_edit.setPlainText(text)

        edit_dialog.exec_()

    
    # Segments -----------------------------------------------------------------

    def set_segment_interaction_buttons_enabled(self, enabled: bool = True):
        """ Enable/disable the buttons that interact with segments. 
    
        Args:    
            enabled (bool): True to enable, False to disable.
        """
        view: SbSWindow = self._view  

        segment_interaction_buttons = (
            view.button_run_segment,
            view.button_copy_segment,
            view.button_remove_segment,
        )
        for button in segment_interaction_buttons:
            button.setEnabled(enabled)
    
    def set_all_segment_buttons_enabled(self, enabled: bool = True):
        """ Enable/disable all segment buttons. 
    
        Args:
            enabled (bool): True to enable, False to disable.
        """
        view: SbSWindow = self._view  

        segment_buttons = (
            view.button_run_segment,
            view.button_copy_segment,
            view.button_remove_segment,
            view.button_new_segment,
            view.button_default_segments,
            view.button_save_segments,
            view.button_load_segments,
        )
        for button in segment_buttons:
            button.setEnabled(enabled)

    @Slot(tuple)
    def segment_selection_changed(self, segments: Sequence[SegmentItemModel] | None = None):
        """ Updates the GUI when the selected segments change.
        If no segments are given, the currently selected segments are used.
        
        Args:
            segments: Sequence[SegmentItemModel]: The new selection of segments.
        """
        view: SbSWindow = self._view  
        self.clear_plots()

        if segments is None:
            segments = view.get_selected_segments()
            
        LOGGER.debug(f"{len(segments)} Segment(s) selected.")
        if not len(segments):
            self.set_segment_interaction_buttons_enabled(False)
            return

        self.set_segment_interaction_buttons_enabled(True)
        self.plot()

    @Slot()
    def add_default_segments(self):
        """ Add default segments to the currently selected measurements. 
        These segments are defined in :data:`omc3_gui.segment_by_segment.defaults.DEFAULT_SEGMENTS`. 
        """
        LOGGER.debug("Adding default segments.")
        view: SbSWindow = self._view  

        selected_measurements = view.get_selected_measurements()
        if not selected_measurements:
            LOGGER.error("Please select at least one measurement.")
            return

        for measurement in selected_measurements:
            self.add_default_segements_to_measurement(measurement)

        self.measurement_selection_changed(selected_measurements)
    
    def add_default_segements_to_measurement(self, measurement: OpticsMeasurement): 
        """ Add default segments to the given measurement. 
        These segments are defined in :data:`omc3_gui.segment_by_segment.defaults.DEFAULT_SEGMENTS`. 

        Args:
            measurement (OpticsMeasurement): The measurement to add segments to.
        """
        if measurement.beam is not None:  # LHC
            for segment_tuple in DEFAULT_SEGMENTS:
                segment = SegmentDataModel(measurement, *segment_tuple)
                segment.start = f"{segment.start}.B{measurement.beam}"
                segment.end = f"{segment.end}.B{measurement.beam}"
                measurement.try_add_segment(segment, silent=True)
            return
        
        # TODO: Implement for other accelerators
        LOGGER.error(f"No beam found in measurement {measurement.display()}. Cannot add default segments.")


    @Slot()
    def new_segment(self):
        """ Create a new segment and add it to the currently selected measurements. """
        LOGGER.debug("Creating new segment.")
        view: SbSWindow = self._view  

        selected_measurements = view.get_selected_measurements()
        if not selected_measurements:
            LOGGER.error("Please select at least one measurement.")
            return
        
        LOGGER.debug("Opening edit dialog for a new segment.")
        dialog = SegmentDialog(parent=view)
        dialog.validate_only_modified = False
        if dialog.exec_() == dialog.Rejected:
            LOGGER.debug("Segment dialog cancelled.")
            return
        
        LOGGER.debug("Segment dialog closed. Updating segement.")
        for measurement in selected_measurements:
            new_segment_copy = dialog.segment.copy()
            new_segment_copy.measurement = measurement
            measurement.try_add_segment(new_segment_copy)

        self.measurement_selection_changed(selected_measurements)
    
    @Slot()
    def copy_segment(self, segments: Sequence[SegmentItemModel] | None = None):
        """ Create a copy of the given segments and add them to the currently selected measurements. 
        If no segments are given, the currently selected segments are copied.

        Args:
            segments: Sequence[SegmentItemModel]: The segments to copy.
        """
        view: SbSWindow = self._view  

        if segments is None:
            segments = view.get_selected_segments()
        
        if not segments:
            LOGGER.error("Please select at least one segment to copy.")
            return

        LOGGER.debug(f"Copying {len(segments)} segments.")
        selected_measurements = view.get_selected_measurements()
        if not selected_measurements:
            LOGGER.error("Please select at least one measurement.")
            return

        for segment_item in segments:
            new_segment_name = f"{segment_item.name}_copy"
            for measurement in selected_measurements:  
                # Check if copied segment name already exists in one of the measurements
                try:
                    measurement.get_segment_by_name(new_segment_name)
                except NameError:
                    pass
                else:
                    LOGGER.error(
                        f"Could not create copy \"{new_segment_name}\" as it already exists in {measurement.display()}."
                    )
                    break
            else:
                # None of the measurements have the copied segment name, so add to the measurements
                for measurement in selected_measurements:
                    for segment in segment_item.segments:
                        new_segment = segment.copy()
                        new_segment.name = new_segment_name
                        new_segment.measurement = measurement
                        measurement.try_add_segment(new_segment)
            
        self.measurement_selection_changed(selected_measurements)
    
    @Slot()
    def remove_segment(self, segments: Sequence[SegmentItemModel] | None = None):
        """ Remove the given segments from the currently selected measurements. 
        If no segments are given, the currently selected segments are removed.
        
        Args:
            segments: Sequence[SegmentItemModel]: The segments to remove.
        """
        view: SbSWindow = self._view  

        if segments is None:
            segments = view.get_selected_segments()
        
        if not segments:
            LOGGER.error("Please select at least one segment to remove.")
            return

        LOGGER.debug(f"Removing {len(segments)} segments.")

        for segment_item in segments:
            for segment_data in segment_item.segments:
                segment_data.measurement.remove_segment(segment_data)
        
        self.measurement_selection_changed(view.get_selected_measurements())

    @Slot()
    def load_segments(self):
        self.load_segments_for_selected_measurements()
        # LOGGER.debug("Loading segments from file/folder.")
        # TODO: implement file saving and loading and ask user which one to do.
    
    def load_segments_for_selected_measurements(self):
        """ Load segments for the currently selected measurements. """
        LOGGER.debug("Loading segments for selected measurements.")
        view: SbSWindow = self._view

        selected_measurements = view.get_selected_measurements()
        if not selected_measurements:
            LOGGER.error("Please select at least one measurement.")
            return

        for measurement in selected_measurements:
            self.load_segments_for_measurement(measurement)

        # Update the view
        self.measurement_selection_changed(selected_measurements)

    def load_segments_for_measurement(self, measurement: OpticsMeasurement):
        """ Load segments for the given measurement. """
        LOGGER.debug(f"Loading segments for {measurement.display()}.")
        
        segments = get_segments_from_directory(measurement.output_dir)
        if not segments:
            LOGGER.debug(f"No segments found in {measurement.output_dir}.")
            return

        for segment_tuple in segments:
            segment = SegmentDataModel(measurement, *segment_tuple)
            measurement.try_add_segment(segment)
        
    @Slot()
    def save_segments(self):
        LOGGER.debug("Saving segments to a file.")
        view: SbSWindow = self._view
        view.showErrorDialog("Error: Not Implemented", "The save segments function is not implemented yet.")
        # TODO
        # Save current segements to a json file

    @Slot()
    def run_segments(self, segments: Sequence[SegmentItemModel] | None = None):
        """ Run the given segments on the currently selected measurements.
        If no segments are given, the currently selected segments are run.
        
        Args:
            segments: Sequence[SegmentItemModel]: The segments to run.
        """
        view: SbSWindow = self._view  

        if segments is None:
            segments: Sequence[SegmentItemModel] = view.get_selected_segments()
        
        if not segments:
            LOGGER.error("Please select at least one segment to run.")
            return

        LOGGER.debug(f"Running {len(segments)} segments.")
        selected_measurements: Sequence[OpticsMeasurement] = view.get_selected_measurements()
        if not selected_measurements:
            LOGGER.error("Please select at least one measurement.")
            return
        
        all_selected_segment_data: list[SegmentDataModel] = [sdata for s in segments for sdata in s.segments]
        measurements_to_run: list[OpticsMeasurement] = [
            meas for meas in selected_measurements if any(s in meas.segments for s in all_selected_segment_data)
        ]

        for idx, measurement in enumerate(measurements_to_run):
            # Filter segments that are in the measurement and sort into segments/elements
            selected_segments_in_meas = [s for s in  all_selected_segment_data if s in measurement.segments]
            segment_parameters = [s.to_input_string() for s in selected_segments_in_meas if not s.is_element()]
            element_parameters = [s.to_input_string() for s in selected_segments_in_meas if s.is_element()] 

            # Create sbs-callable from measurement/inputs
            sbs_function  = partial(
                    segment_by_segment, 
                    **measurement.get_sbs_parameters(),
                    segments=segment_parameters or None,
                    elements=element_parameters or None,
                )

            # Create thread
            measurement_task = BackgroundThread(
                function=sbs_function,
                message=f"SbS for {measurement.display()}",
            )
            
            # For Real Use: Run Task ---
            LOGGER.info(f"Starting {measurement_task.message}")
            self._add_running_task(task=measurement_task)
            measurement_task.start()
            # -------------------------------------

            # For Debugging: Start sbs directly ---
            # sbs_function()
            # self.clear_all_data()
            # LOGGER.info(f"Finished {measurement_task.message}")
            # -------------------------------------
    
    @Slot()
    def clear_all_data(self):
        """ Clear all segment data on all measurements. """
        view: SbSWindow = self._view
        measurements = view.get_all_measurements()
        for meas in measurements:
            clear_segments(meas.segments)
        self.plot()  # update plots -> reads needed data new

# Plotting ---------------------------------------------------------------------
    def plot(self, fail_ok: bool = True):
        """ Trigger a plot update with the currently selected segments. 
        
        This function is called when the user changes the selected segments or the settings.
        As a segment selection change is often triggered, i.e. when the user clicks on a segment in the table,
        e.g. to actually run this segment, one should not consider that the plot NEEDS to be updated.
        Hence in the default settings with ``fail_ok=True``, this function will mostly log to debug.

        This function is also called after the plotting settings are changed 
        (either in the settings dialog or in the menu), hence this is also a good place to warn the user
        if he did some mistakes there.
        """
        log_function = LOGGER.debug if fail_ok else LOGGER.error

        view: SbSWindow = self._view
        settings: PlotSettings = self.settings.plotting
        definition, widget = view.get_current_tab()
        
        self.clear_plots()

        if not settings.forward and not settings.backward:
            LOGGER.error("Please enable at least one propagation method to show.")
            return

        widget.set_connect_x(settings.connect_x)
        widget.set_connect_y(settings.connect_y)

        segments = view.get_selected_segments()
        if not len(segments):
            log_function("Not plotting, no segments selected.")
            return

        segments_data: list[SegmentDataModel] = [s_data for s in segments for s_data in s.segments if s_data.has_run()]
        if not len(segments_data):
            log_function("Not plotting, no segments have been run.")
            return

        if settings.same_start and not settings.model_s:  # only an issue if they all start at 0
            starts = {re.sub(r"\.B[12]$", "", s.start, flags=re.IGNORECASE) for s in segments_data}
            if len(starts) > 1:
                log_function("Not plotting, segments have different start BPMs (see 'Same Start' in settings).")
                return
        
        plot_segment_data(
            widget=widget, 
            definitions=definition, 
            segments=segments_data, 
            settings=settings,
        )

    def clear_plots(self):
        """ Clear the plots. """
        view: SbSWindow = self._view
        widget: DualPlotWidget = view.get_current_tab()[1]
        widget.clear()

# Other ------------------------------------------------------------------------
    @Slot()
    def show_settings(self):
        LOGGER.debug("Showing settings.")
        settings_dialog = SettingsDialog(settings=self.settings)
        if settings_dialog.exec_():
            view: SbSWindow = self._view
            view.update_menu_settings(
                menu="View",
                settings=self.settings.plotting,
            )
            self.plot()
        

# Helper Functions ------------------------------------------------------------

def clear_segments(segments: Sequence[SegmentDataModel]):
    """ Clear all chached segment data, so that the GUI loads the new SbS data. """
    for segment in segments:
        segment.data.clear()



""" 
UI: Styles
----------

Helper functions to style UI elements and plots.
"""
from omc3_gui.ui_components import colors

MONOSPACED_TOOLTIP = f"""
    QToolTip {{
        background-color: {colors.TOOLTIP_BACKGROUND}; /* Light gray background */
        color: {colors.TOOLTIP_TEXT}; /* Dark gray text */
        border: 1px solid {colors.TOOLTIP_BORDER}; /* Gray border */
        font-family: "Courier New", monospace; /* Monospaced font */
    }}
"""


""" 
Help Dialogs
------------
"""
from __future__ import annotations

from qtpy.QtCore import Qt
from qtpy.QtWidgets import QMessageBox

def show_help_dialog():
    """ Displays the help dialog for the segment-by-segment GUI. """

    help_text = """
    <h3> Frequently Asked Questions </h3><br>
    
    <br>
    <b> How do I open a Measurement?</b><br>
    
    One way to open measurements automatically, is to give them as command line
    arguments when starting the sbs_gui, either <i> -m </i> or <i> --measurements </i>.<br>
    If you want to load them manually, you click the <i>Load</i> button.<br><br>

    The given/selected measurements can be either <i>omc3 optics folders, folders containing sbs-json files
    or sbs-json files</i> directly.
    The latter are created automatically in the measurement-output folder when editing a loaded measurement.<br>

    <br>
    
    <b> Do I have to invert my corrections when using them in the machine?</b><br>

    YES! <i>(but it depends)</i><br>
    The "corrections" here are actually used to match the model opttics to the 
    mesured optics (see the info about the dashed "corr" line below). 
    Therefore you have to invert them, to actually use them as corrections in the machine. <br>
    <b> NOTE </b> that these are the MAD-X values. Make how the signs are actually
    mapped in the machine! <br>

    <br>

    <b> What is the solid line?</b><br>

    The solid line is the difference between the Measurement and 
    the propagated model, i.e. the Measurement at the start (or end) of the segment 
    propagated through the nominal model via MAD-X.<br>
    This line therefore shows you how much the optics deviate through the segment 
    from the nominal model.<br> 

    <br>   
    
    <b> What is the dashed line that says "corr"?</b><br>

    This is the difference between the <i>"corrected"</i> propagated model and the 
    nominal propagated model.<br>
    This means in both cases the measured values are used as initial conditions.
    What you are trying to achieve is a match between the dashed and the solid line,
    because that means that now your model matches the optics in the measured data.<br>

    <br>
    
    <b> What is the dashed line that says "expct"?</b><br>

    This is the difference between the Measurement and the "corrected" propagated model
    and is therefore the <i>expected</i> measured difference to the nominal model after 
    applying the correction in the machine (same as in global correction).<br>
    You can activate this view via the plot-settings <i>"Expectation"</i>.<br>

    <br>

    <b>Shortcuts</b><br>

    In Graph:<br>
    <i>Double-Click</i> : Zoom history back one step. <br>
    <i>Right-Click</i> : Zoom history back one step. <br>
    <i>Shift + Right-Click</i> : Zoom history back all steps. <br>
    <i>Alt + Right-Click</i> : pyqtgraph context menu. <br>

    <br>

    In Measurements-List:<br>
    <i>Double-Click</i> : Edit the Measurement.<br>

    <br> 
    
    """
    msg_box = QMessageBox(icon=QMessageBox.Information)
    msg_box.setWindowTitle("Help")
    msg_box.setTextFormat(Qt.TextFormat.RichText)
    msg_box.setText(help_text)
    msg_box.exec_()



""" 
UI: Text Editor Dialog
----------------------

This module provides a dialog for editing text files.
"""

from pathlib import Path

from qtpy.QtWidgets import QDialog, QDialogButtonBox, QTextEdit, QVBoxLayout

from omc3_gui.ui_components.message_boxes import show_error_dialog 

class TextEditorDialog(QDialog):
    def __init__(self, file_path: str | Path, title: str = "Edit {}", parent = None):
        super().__init__(parent)
        self.file_path: Path = Path(file_path)
        self.setWindowTitle(title.format(self.file_path.name))  # maybe better absolute path?

        self.text_edit = QTextEdit(self)
        self.load_file_content()

        self.buttons = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel, self)
        self.buttons.accepted.connect(self.save_changes)
        self.buttons.rejected.connect(self.reject)

        layout = QVBoxLayout()
        layout.addWidget(self.text_edit)
        layout.addWidget(self.buttons)
        self.setLayout(layout)

    def load_file_content(self):
        try:
            content = self.file_path.read_text()
        except FileNotFoundError:
            return  # empty or new file

        self.text_edit.setPlainText(content)

    def save_changes(self):
        content = self.text_edit.toPlainText()
        try:
            self.file_path.write_text(content)
        except PermissionError:
            show_error_dialog(
                message=f"Could not write to file {self.file_path.absolute()}.",
                title="Permission Error",
                parent=self,
            )
            return
            
        self.accept()


""" 
DataClass UI: Controller
------------------------

This module provides classes to generate simple UI's for dataclasses,
which allow to edit the values of dataclasses in an easy way.

Use these controllers to create and steer the UI's.
"""
from __future__ import annotations

import dataclasses as dc
import logging
from dataclasses import Field, dataclass, fields
from functools import partial
from pathlib import Path

from qtpy import QtWidgets

from omc3_gui.ui_components import colors, file_dialogs
from omc3_gui.ui_components.dataclass_ui import view, model
from omc3_gui.ui_components.dataclass_ui.model import FieldUIDef, FilePath, DirectoryPath  # noqa: F401 paths need to be known by `get_dataclass_types`  
from omc3_gui.ui_components.widgets import HorizontalSeparator
from typing import get_type_hints, TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Callable, Sequence

LOGGER = logging.getLogger(__name__)


@dataclass()
class FieldUI:
    """ Controller for a single field in a dataclass, connecting it to a widget.
    
    This class is not meant to be used directly, but is rather automatically
    generated by and attached to :class:`omc3_gui.utils.dataclass_ui.controller.DataClassUI`.
    """
    widget: QtWidgets.QWidget  # edit-widget to edit field value
    label: QtWidgets.QLabel    # label-widget of the field
    get_value: Callable        # getter for the widget value, returns the value as appropriate type for the dataclass
    set_value: Callable        # setter for the widget value
    text_color: str | None = colors.TEXT_DARK  # default text-color for both widget and label
    modified: bool = False     # flag indicating if the widget-content has been modified by the user

    def __post_init__(self):
        """ Connects the widget to the label and sets the text-color. """
        for main_attribute in ["text", "value", "state"]:
            try:
                signal = getattr(self.widget, f"{main_attribute}Changed")
            except AttributeError:
                pass
            else:
                signal.connect(self.has_changed)
                break
        self.widget.setStyleSheet(f"color: {self.text_color};")
        self.label.setStyleSheet(f"QLabel {{color: {self.text_color}}};") 

    def has_changed(self):
        """ Triggered when the widget has been modified.
        Sets then the modified flag and changes label font.
         """
        self.modified = True
        font = self.label.font()
        font.setItalic(True)
        self.label.setFont(font)
    
    def reset(self):
        """ Reset the label font to normal and clear the modified flag. """
        self.modified = False
        font = self.label.font()
        font.setItalic(False)
        self.label.setFont(font)


class DataClassUI:
    """ Controller for the UI representation of a dataclass. 
    It contains a grid-layout that can be added to any QWidget/QLayout.   
    """

    def __init__(self, 
                field_definitions: Sequence[FieldUIDef], 
                dclass: type | object, 
                layout: QtWidgets.QLayout | None = None
        ):
        """ Builds a DataClassUI from a list of field definitions.

        NOTE: `dclass` is not automatically attached to the resulting class, unless it is an already an instance.
        This function works with classes and instances, but the attached object needs to be the instance.
        
        Args:
            field_def (Sequence[FieldUIDef | str]): list of field definitions
            dclass (type | object): DataClass type or instance.
            layout (QtWidgets.QLayout | None): Layout to add the UI to. If None, a new grid-layout is created.

        Returns:
            DataClassUI: An instances with a grid-layout containing edit-widgets and labels.
        """
        if layout is None:
            layout = QtWidgets.QGridLayout()
        
        self.model: object = None if isinstance(dclass, type) else dclass  # dataclass instance 
        self.layout: QtWidgets.QGridLayout = layout  # final layout of the UI for dataclass
        self.fields: dict[str, FieldUI] = {} # stored field UI-elements

        field_instances: dict[str, Field] = {field.name: field for field in fields(dclass)}
        field_types = get_dataclass_types(
            dclass, 
            [
                field.name for field in field_definitions 
                if field is not None and not isinstance(field, str) and field.name in field_instances.keys()
             ]
        )
        # 
        
        for idx_row, field_def in enumerate(field_definitions):
            if field_def is None:  # Separator ---
                layout.addWidget(HorizontalSeparator(), idx_row, 0, 1, 3)
                continue

            if isinstance(field_def, str):  # Label only ---
                layout.addWidget(QtWidgets.QLabel(field_def), idx_row, 0)
                continue
            
            if field_def.name not in field_instances:
                raise ValueError(f"Field {field_def.name} not found in dataclass {dclass}")

            field_inst = field_instances[field_def.name]
            
            # Label ---
            qlabel = QtWidgets.QLabel(field_def.label or field_inst.metadata.get("label", field_def.name))
            qlabel.setToolTip(field_def.comment or field_inst.metadata.get("comment", ""))
            layout.addWidget(qlabel, idx_row, 0)
            
            # User input ---
            # If field_def.type is not given, use evaluate from dataclass. 
            # Check __args__ in case of Union/Optional and use first one.
            # The type needs to be instanciable!
            eval_type = field_types[field_def.name]  # evaluated type, might be Optional/Union etc with __args__
            field_type = field_def.pytype or getattr(eval_type, "__args__", [eval_type])[0]  

            widget = TYPE_TO_WIDGET_MAP.get(field_type, QtWidgets.QLineEdit)()

            try:
                widget.setReadOnly(not field_def.editable)
            except AttributeError:
                widget.setEnabled(field_def.editable)

            get_value, set_value = build_getter_setter(widget, field_type)
            self.fields[field_def.name] = FieldUI(
                widget=widget, 
                label=qlabel,
                set_value=set_value,
                get_value=get_value,
                text_color=colors.TEXT_DARK if field_def.editable else colors.GREYED_OUT_TEXT_DARK
            )
            
            if not issubclass(field_type, Path) or not field_def.editable:
                layout.addWidget(widget, idx_row, 1, 1, 2)
            else:
                layout.addWidget(widget, idx_row, 1)
                # Add Path selection button ---
                button = QtWidgets.QPushButton("...")
                button.setFixedWidth(30)
                
                if field_type is model.FilePath:
                    dialog = file_dialogs.OpenFileDialog
                elif field_type is model.DirectoryPath:
                    dialog = file_dialogs.OpenDirectoryDialog 
                else:
                    dialog = file_dialogs.OpenAnySingleDialog

                button.clicked.connect(partial(
                    run_dialog, 
                    dialog=dialog,
                    get_value=get_value,
                    set_value=set_value))
                layout.addWidget(button, idx_row, 2)

        if self.model is not None:
            self.update_ui()  # fill with values from the instance

    def reset_labels(self):
        """ Resets all labels to indicate that the field shows the currently set value in the dataclass."""
        for name in self.fields.keys():
            self.fields[name].reset()
    
    def update_widget_from_model(self, name: str):
        """ Updates the edit-widget of the given field from the dataclass values. 
        
        Args:
            name (str): name of the field in the dataclass
        """
        if self.model is None:
            raise ValueError("No dataclass instance attached to the UI.")

        value = getattr(self.model, name)
        if value is not None:
            self.fields[name].set_value(value)
    
    def update_model_from_widget(self, name: str):
        """ Updates the dataclass value of the given field from the edit-widget. 
        
        Args:
            name (str): name of the field in the dataclass
        """
        if self.model is None:
            raise ValueError("No dataclass instance attached to the UI.")

        field: FieldUI = self.fields[name]

        if not field.modified:  # avoid replacing 'None' with widget defaults
            LOGGER.debug(f"Field {name} was not modified.")
            return 

        value = field.get_value()
        setattr(self.model, name, value)

    def update_ui(self):
        """ Updates all edit-widgets from the dataclass values. """
        for name in self.fields.keys():
            self.update_widget_from_model(name)

    def update_model(self):
        """ Updates all dataclass fields from the current edit-widget values. """
        for name in self.fields.keys():
            self.update_model_from_widget(name)

    def validate(self, only_modified: bool = False):
        """ Checks all edit-widgets for valid choices. 
        
        The validation function NEED to return a truthy value, if 
        the choice is valid. If it is invalid, they can either return a falsy
        value or raise an exception. In the latter case, this exception will be
        printed instead of the default message.
        """
        if self.model is None:
            raise ValueError("No dataclass instance attached to the UI.")

        invalid_fields_str = []
        for name in self.fields.keys():
            field = self.model.__dataclass_fields__[name]
            field_ui = self.fields[field.name]
            if only_modified and not field_ui.modified:
                continue
            
            validate_function = field.metadata.get("validate")
            if validate_function is not None:
                value = self.fields[field.name].get_value()

                validation_result = False
                try: 
                    validation_result = validate_function(value)
                except ValueError as e:
                    invalid_fields_str.append(f"{field_ui.label.text()}: {str(e)}")
                else: 
                    if not validation_result:
                        invalid_fields_str.append(f"{field_ui.label.text()}: {value} is not a valid choice.")

        if invalid_fields_str:
            full_str = "\n".join(invalid_fields_str)
            raise ValueError(f'The following fields contain wrong values:\n{full_str}')


class DataClassTabbedUI():
    """ Controller for the UI representation of a dataclass, containing other dataclasses as fields. 
    It contains a tabbed layout that can be added to any QWidget.   
    """

    def __init__(self, 
                dclass: object | type, 
                layout: QtWidgets.QLayout  | None = None, 
                widget: QtWidgets.QTabWidget | None = None
        ):
        """ Builds a DataClassTabbedUI from a dataclass, using DataClassUIs as widgets in the tabs.
        NOTE: As with the DataClassUI, the dataclass (`dclass`) is not automatically
              attached to the resulting class, 
        
        Args:
            dclass (object | type): dataclass instance or class. 
                                    Needs to contain dataclasses as fields, which will converted to tabs.
            layout (QtWidgets.QLayout | None): Layout to add the UI widget to. 
                                               If None, a new QVBoxLayout is created.
            widget (QtWidgets.QTabWidget | None): TabWidget to add the UI tabs to.
                                                  If None, a new QTabWidget is created.
                                                  The widget is always added to the layout.
        """
        self.layout: QtWidgets.QLayout = layout or QtWidgets.QVBoxLayout()
        self.widget: QtWidgets.QTabWidget = widget or QtWidgets.QTabWidget()
        self._dataclass_uis: dict[str, DataClassUI] = {}

        self.layout.addWidget(self.widget)

        fields = [field.name for field in dc.fields(dclass) if field.name[0] != "_"]
        for field_name in fields:
            sub_dclass = getattr(dclass, field_name)
            dataclass_ui = DataClassUI(
                field_definitions=[FieldUIDef(field.name) for field in dc.fields(sub_dclass) if field.name[0] != "_"],
                dclass=sub_dclass,
            )
            ui_widget = QtWidgets.QWidget()
            ui_widget.setLayout(dataclass_ui.layout)
            self.widget.addTab(ui_widget, field_name.capitalize())
            self._dataclass_uis[field_name] = dataclass_ui
        
        self._model: object = None
        if not isinstance(dclass, type):
            self.model = dclass  # also updates ui with values
        
    @property
    def model(self):
        return self._model
    
    @model.setter
    def model(self, model: object):
        """ Sets the model and updates the dataclass_uis. 
        This does not add new dataclass_uis to the layout and throws errors 
        if fields in the model are missing. """
        self._model = model
        for name, dataclass_ui in self._dataclass_uis.items():
            dataclass_ui.model = getattr(model, name)
            dataclass_ui.update_ui()

    # Function to pass on to dataclass_uis ---
    def update_ui(self):
        for dataclass_ui in self._dataclass_uis.values():
            dataclass_ui.update_ui()

    def update_model(self):
        for dataclass_ui in self._dataclass_uis.values():
            dataclass_ui.update_model()
    
    def validate(self, only_modified: bool = False):
        for dataclass_ui in self._dataclass_uis.values():
            dataclass_ui.validate(only_modified=only_modified)
    
    def reset_labels(self):
        for dataclass_ui in self._dataclass_uis.values():
            dataclass_ui.reset_labels()


# Helpers for getter/setter and widget type ------------------------------------

TYPE_TO_WIDGET_MAP = {
    int: view.QFullIntSpinBox,  # Maybe just use QLineEdit as well?
    float: QtWidgets.QLineEdit,
    str: QtWidgets.QLineEdit,
    Path: QtWidgets.QLineEdit,
    bool: QtWidgets.QCheckBox,
}


def build_getter_setter(widget: QtWidgets.QWidget, field_type: type) -> tuple[Callable, Callable]:
    """ Getter/Setter Factory for widgets. 
    
    Args:
        widget (QtWidgets.QWidget): The widget to get/set the value of.
        field_type (type): The type of the dataclass field.
    """
    if isinstance(widget, QtWidgets.QCheckBox):
        def get_value() -> bool:
            return widget.isChecked()

        def set_value(value: bool):
            widget.setChecked(value)
    
    elif isinstance(widget, QtWidgets.QSpinBox):
        def get_value():
            return field_type(widget.value())

        def set_value(value):
            widget.setValue(value)
        
    else:   # Any kind of widget should be able to handle strings.      
        def get_value():
            if not widget.text():
                return None
            return field_type(widget.text())

        def set_value(value):
            widget.setText(str(value))
    return get_value, set_value        


def run_dialog(dialog: file_dialogs.OpenFilesDialog, get_value: Callable, set_value: Callable):
    """ Asks the user to select a directory/file, using the `get_value` function 
    o determine the default directory and passing the result to `set_value` if not `None`.

    Args:
        dialog (file_dialogs.OpenFilesDialog): The dialog to run.
        get_value (Callable): The function to get the last value, e.g. from a widget.
        set_value (Callable): The function to set the value returned from the dialog.
    """
    old_path: Path = get_value()
    if old_path is not None:
        old_path.parent 
    else:
        old_path = Path() 
    path = dialog(directory=old_path).run_selection_dialog()
    if path is not None:
        set_value(path)


def get_dataclass_types(dclass: type | object, names: Sequence[str]):
    """ 
    Returns a dictionary mapping field names to their associated types.

    Why this weird way?
    First of all, we avoided the cyclic imports of type-hints by using 
    the `from __future__ import annotations`, but whith 
    this, all type-hints on the dataclasses become `ForwardReferences`, 
    i.e. strings. So if we get them from the dataclass-fields
    via `field.pytype` or `field.pytype.__args__`, we cannot instanciate them later on.

    `get_type_hints()` from `typing` solves this problem, but it 
    evaluates the type-hints of all fields in the dataclass. 
    But this can cause again import errors, because the classes were not actually imported
    (this is why there are no cyclic imports)!

    So this function allows using `get_type_hints()` on the dataclass 
    itself, but only gets the types of the fields that we need, in the hopes 
    that there is no import problem with those.
    
    In the use-case of the DataClass UI here, we should only have very basic 
    types for the fields, so this should be fine.
    
    If you are here because there is, you need to redesign your dataclass/imports.
    Sorry.

    Args:
        dclass (type): The data class to extract field types from.
        names (Sequence[str]): The names of the fields to extract types from.

    Returns:
        Dict[str, type]: A dictionary mapping field names to their associated types. 
    """
    required_annotations = {
        name: value for name, value in dict(dclass.__annotations__).items() if name in names
    }
    class SafeMock:
        __annotations__ = required_annotations
    return get_type_hints(SafeMock)



""" 
UI: Item Models
---------------

Classes that makes it easier to handle unique items in UI models.
"""
from __future__ import annotations

from typing import Any, Protocol
from collections.abc import Hashable, Sequence

from qtpy import QtCore
from qtpy.QtCore import Qt


class Item(Protocol):
    """ Protocol for a class that has an 'id'-property. 
    As used in :class:`omc3_gui.segment_by_segment.main_model.UniqueItemListModel`,
    this id defines whether two items are "the same", i.e. only one of them 
    can be present in the model.
    Example: For the Segments, this should be the name, as if we have two segements
    with the same name, running them overwrites each other.
    """
    id: Hashable


class UniqueItemListModel:
    """ Mixin-Class for a class that has a dictionary of items. 
    Note: I have considered using QAbstractItemModel/QStandardItemModel, 
    but I do not need all the features it provides, so this should be easier
    and allows for keeping items unique (jdilly, 2023).
    All items need to have an 'id'-property.
    """
    def __init__(self):
        self._items: list[Item] = []

    def try_emit_change(self, emit: bool = True):
        """ Emits a dataChanged-signal if the model has changed, and if the 
        class provides such a signal. """
        if not emit:
            return

        if hasattr(self, "dataChanged"):
            # TODO: return which data has actually changed?
            try:
                idx_start = self.index(0)  # list
                idx_end = self.index(self.rowCount()-1)
            except TypeError:
                idx_start = self.index(0, 0)  # table
                idx_end = self.index(self.rowCount()-1, self.columnCount()-1)
                self.headerDataChanged.emit(Qt.Horizontal, 0, self.columnCount()-1)
            self.dataChanged.emit(idx_start, idx_end)
    
    def add_item(self, item: Item, emit: bool = True):
        """ Adds an item to the model, 
        if an item with the same id is not already present. """
        if item.id in [i.id for i in self._items]:
            raise ValueError(f"Item {item.id} already exists")
        self._items.append(item)
        self.try_emit_change(emit)

    def add_items(self, items: Sequence[Item]):
        """ Adds all items from a list to the model, 
        for which items with the same id are not already present. """
        already_exist_items = []
        for item in items:
            try:
                self.add_item(item, emit=False)
            except ValueError:
                already_exist_items.append(item)
        self.try_emit_change()
        if already_exist_items:
            raise ValueError(f"Items already exist: {already_exist_items}")

    def remove_item(self, item: Item, emit: bool = True):
        """ Removes an item from the model. """
        self._items.remove(item)
        self.try_emit_change(emit)
    
    def remove_items(self, items: Sequence[Item]):
        """ Removes all items from a list from the model, if they exist. """
        do_not_exist_items = []
        for item in items:
            try:
                self.remove_item(item, emit=False)
            except ValueError:
                do_not_exist_items.append(item)
        self.try_emit_change()
        if do_not_exist_items:
            raise ValueError(f"Items do not exist: {do_not_exist_items}")
    
    def clear(self):
        """ Removes all items from the model. """
        self._items = []
        self.try_emit_change()

    def remove_item_at(self, index: int):
        self.remove_item(self.get_item_at(index))

    def remove_items_at(self, indices: Sequence):
        self.remove_items([self.get_item_at(index) for index in indices])

    def get_item_at(self, index: int) -> Any:
        return self._items[index]

    def get_index(self, item: Item) -> QtCore.QModelIndex:
        idx_item = self._items.index(item)
        try:
            return self.index(idx_item)
        except TypeError:
            return self.index(idx_item, 0)



""" 
Segment Model
-------------

This module contains the model for the Segments 
in the Segment-by-Segment application.
"""
from __future__ import annotations

from collections import namedtuple
from dataclasses import dataclass
from pathlib import Path
import re
from typing import TYPE_CHECKING

from omc3.segment_by_segment.segments import SegmentDiffs, EXT
from omc3.optics_measurements.constants import NAME
import tfs

from omc3_gui.ui_components import colors
from omc3_gui.ui_components.dataclass_ui import metafield
from omc3_gui.ui_components.item_models import Item

if TYPE_CHECKING:
    from omc3_gui.segment_by_segment.measurement_model import OpticsMeasurement

OK: str = f"<font color=\"{colors.GREEN_DARK}\">✓</font>"
NO: str = f"<font color=\"{colors.RED_DARK}\">✗</font>"
TO_BE_DEFINED: str = "to_be_defined"


def not_empty(value: str | None) -> bool:
    if value is None:
        return False
    value = value.strip()
    return value != "" and value != TO_BE_DEFINED


SegmentTuple = namedtuple('Segment', ['name', 'start', 'end'])  # the simplest way to store a segment definition


@dataclass(slots=True)
class SegmentDataModel:
    """" Container for the segment data, which is also used in the Segment creation dialog. """

    measurement: OpticsMeasurement
    name: str =         metafield("Name",  "Name of the Segment", default=TO_BE_DEFINED, validate=not_empty)
    start: str | None = metafield("Start", "Start of the Segment", default=None, validate=not_empty)
    end: str | None =   metafield("End",   "End of the Segment",   default=None, validate=not_empty)
    _data: SegmentDiffs | None = None

    def __str__(self):
        return self.name

    def is_element(self):
        return is_element(self)

    def to_input_string(self):
        """ String representation of the segment as used in inputs."""
        return to_input_string(self)
    
    @property
    def data(self) -> SegmentDiffs:
        if self._data is None or self._data.directory != self.measurement.output_dir or self._data.segment_name != self.name:
            self._data = SegmentDiffs(self.measurement.output_dir, self.name)
        return self._data
    
    def has_run(self) -> bool:
        try:
            return self.data.get_path("phase_x").is_file()
        except AttributeError:
            return False
        # TODO: Maybe load and check first and last BPM? (jdilly, 2025)

    def clear_data(self):
        self._data = None
    
    def copy(self):
        return SegmentDataModel(measurement=self.measurement, name=self.name, start=self.start, end=self.end)


class SegmentItemModel(Item):
    """ Model for a segment item in the Segment-Table of the Segment-by-Segment application. 
    Each item has name, start and end and attached a list of actual segment-obejcts
    """

    def __init__(self, name: str, start: str = None, end: str = None):
        self._name = name
        self._start = start
        self._end = end
        self._segments: list[SegmentDataModel] = []

    @classmethod
    def from_segments(cls, segments: list[SegmentDataModel]) -> SegmentItemModel:
        new = cls(segments[0].name, segments[0].start, segments[0].end)
        new.segments = segments  # also checks for equality of given segments
        return new
    
    @classmethod
    def from_segment(cls, segment: SegmentDataModel) -> SegmentItemModel:
        new = cls(segment.name, segment.start, segment.end)
        return new

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value: str):
        self._name = value
        for segment in self.segments:
            segment.name = value
    
    @property
    def start(self) -> str:
        return self._start

    @start.setter
    def start(self, value: str):
        self._start = value
        for segment in self.segments:
            segment.start = value

    @property
    def end(self) -> str:
        return self._end

    @end.setter
    def end(self, value: str):
        self._end = value
        for segment in self.segments:
            segment.end = value

    @property
    def segments(self) -> list[SegmentDataModel]:
        return self._segments
    
    @segments.setter
    def segments(self, segments: list[SegmentDataModel]):
        if any(not compare_segments(self, segment) for segment in segments):
            raise ValueError(
                "At least one given segment has a different "
                f"definition than the others or than this {self.__class__.name}."
            )
        self._segments = segments

    def append_segment(self, segment: SegmentDataModel):
        if not compare_segments(self, segment):
            raise ValueError(f"Given segment has a different definition than this {self.__class__.name}.")
        self.segments.append(segment)

    @property
    def id(self) -> str:
        """ Unique identifier for the segment. """
        return self.name + self.start + self.end
    
    def tooltip(self) -> str:
        """ Returns a string with information about the segment, 
        as to be used in a tool-tip.  
        
        Hint: Use fully HTML compatible strings, otherwise Qt will revert to plain text. 
              e.g. use <br> instead of \\n. Also &nbsp; instead of whitespaces, as they are collapsed in HTML.
    
        """
        parts = [
            f"<tr><td>{OK if segment.has_run() else NO}</td>"
            "<td></td>"
            f"<td>{segment.measurement.display()}</td></tr>" 
            for segment in self.segments
        ]
        return "<tr><th>Run</th><th>|</th><th>In Measurement</th></tr>" + "".join(parts)

    def is_element(self):
        return is_element(self)

    def to_input_string(self):
        """ String representation of the segment as used in inputs."""
        return to_input_string(self)


# Segment functions ---

def compare_segments(a: SegmentDataModel | SegmentItemModel, b: SegmentDataModel | SegmentItemModel) -> bool:
    """ Checks if two Segments have the same definition. """
    return a.name == b.name and a.start == b.start and a.end == b.end


def is_element(segment: SegmentItemModel | SegmentDataModel) -> bool:
    """ Checks if the segment is an element-segment. """
    return segment.start is None or segment.end is None


def to_input_string(segment: SegmentItemModel | SegmentDataModel) -> str:
    """ Convert the segment to the string representation as used in inputs. """
    if is_element(segment):
        return segment.name
    return f"{segment.name},{segment.start},{segment.end}"


# Other ---

def get_segments_from_directory(directory: Path) -> list[SegmentTuple]:
    """ Parse segments from a directory. 
    
    This function needs to be kept synchronized with :class:`omc3.segment_by_segment.segments.SegmentDiffs`.
    """
    pattern = fr"{SegmentDiffs.PREFIX}.+?_[xy]_(.+)\{EXT}"  # keep in sync, hint: '.' in EXT
    segments: list[SegmentTuple] = []
    names: set[str] = set()
    for file in directory.glob(f"{SegmentDiffs.PREFIX}*{EXT}"):
        file_match = re.match(pattern, file.name)
        if file_match is None:
            continue

        name = file_match.group(1)
        if name not in names:
            names.add(name)
            df = tfs.read(file, index=NAME)
            segments.append(SegmentTuple(name, df.index[0], df.index[-1]))
    return segments


""" 
Segment-by-Segment GUI
----------------------

Graphical user interface to run the Segment-by-Segment propagation.
"""
from argparse import ArgumentParser
from pathlib import Path
import sys
from omc3_gui.segment_by_segment.main_controller import SbSController
from omc3_gui.segment_by_segment.settings import Settings
from omc3_gui.utils.log_handler import init_logging

# --- For QT Debugging ----------------
# import os
# os.environ["QT_DEBUG_PLUGINS"] = "1"
# -------------------------------------


def parse_args() -> tuple[list[Path], Settings]:
    args = ArgumentParser()
    args.add_argument(
        "-m",
        "--measurements",
        nargs="+",
        type=Path,
        help="Measurements to process",
    )
    args.add_argument(
        "-s",
        "--settings",
        type=Path,
        help="Settings file to use",
    )
    opt = args.parse_args()
    return opt.measurements,  opt.settings


if __name__ == "__main__":
    init_logging()
    measurements, settings = parse_args()
    sys.exit(SbSController.run_application(
        measurements=measurements, 
        settings=settings
    ))


""" 
Plotting
--------

Plots for segment-by-segment.
"""
from __future__ import annotations

from dataclasses import dataclass
from functools import cache
import logging
from pathlib import Path

from omc3.definitions.optics import (
    S_COLUMN,
    S_MODEL_COLUMN,
    RDT_AMPLITUDE_COLUMN, 
    RDT_PHASE_COLUMN,
    RDT_REAL_COLUMN,
    RDT_IMAG_COLUMN,
    ColumnsAndLabels,
)
from omc3.segment_by_segment.propagables import PropagableColumns
from omc3.model.constants import TWISS_ELEMENTS_DAT
from omc3.optics_measurements.constants import NAME
from qtpy.QtCore import Qt
import tfs

from omc3_gui.plotting.classes import DualPlotWidget
from omc3_gui.plotting.element_lines import plot_element_lines
from omc3_gui.plotting.latex_to_html import latex_to_html_converter
from omc3_gui.plotting.tfs_plotter import plot_dataframes
from omc3_gui.segment_by_segment.segment_model import SegmentDataModel
from omc3_gui.segment_by_segment.settings import PlotSettings

LOGGER = logging.getLogger(__name__)


PenStyle = Qt.PenStyle

@dataclass(frozen=True)
class PlotDefinition:
    file_name: str
    ylabel: str
    column: str 
    error_column: str
    propagable_columns: PropagableColumns

    @classmethod
    def create(cls, file_name: str, columns: ColumnsAndLabels):
        return cls(
                file_name=file_name,
                ylabel=latex_to_html_converter(columns.delta_label),
                column=columns.column,
                error_column=columns.error_column,
                propagable_columns=PropagableColumns(columns.column, plane="")
            
        )


@dataclass(frozen=True)
class DualPlotDefinition:
    name: str
    top: PlotDefinition
    bottom: PlotDefinition

    @property
    def plots(self) -> tuple[PlotDefinition, PlotDefinition]:
        return (self.top, self.bottom)

    @classmethod
    def generate_xy(cls, name: str, file_name: str, columns: ColumnsAndLabels):
        """ Generate a DualPlotDefinition for XY-Planed Plots. """
        return cls(name, *(
            PlotDefinition.create(f"{file_name}_{plane}", columns.set_plane(plane.upper())) for plane in "xy"
        ))
    
    @classmethod
    def generate_amplitude_phase(cls, file_name: str):
        """ Generate a DualPlotDefinition for Amplitude/Phase Plots. """
        name = f"{file_name} A/φ"
        amp = RDT_AMPLITUDE_COLUMN.set_label_formatted(file_name)
        phase = RDT_PHASE_COLUMN.set_label_formatted(file_name)
        return cls.generate_rdt(name, file_name, amp, phase)
    
    @classmethod
    def generate_real_imag(cls, file_name: str):
        """ Generate a DualPlotDefinition for Amplitude/Phase Plots. """
        name = f"{file_name} Re/Im"
        real = RDT_REAL_COLUMN.set_label_formatted(file_name)
        imag = RDT_IMAG_COLUMN.set_label_formatted(file_name)
        return cls.generate_rdt(name, file_name, real, imag)

    @classmethod
    def generate_rdt(cls, name: str, file_name: str,  columns_top: ColumnsAndLabels, columns_bottom: ColumnsAndLabels):
        """ Generate a DualPlotDefintion for RDTs, i.e. Amplitude/Phase, Re/Im. """
        return cls(name, *(
            PlotDefinition.create(file_name, columns) for columns in (columns_top, columns_bottom)
        ))


class DirectionStyle:
    """ Helper Class to define the Style based on direction and expected. """

    def __init__(self, direction: str, expected: bool | None):
        self.direction = direction
        self.expected = expected

    @property
    def linestyle(self):
        if self.expected is None:
            return PenStyle.SolidLine
        return PenStyle.DashLine

    @property
    def brightness(self):
        return {
            "forward": None,
            "backward": 150
        }[self.direction]

    @property
    def marker(self):
        return {
            "forward": "t2",
            "backward": "t3"
        }[self.direction]

    @property
    def suffix(self):
        shorthand = {
            "forward": " fwd",
            "backward": " bwd"
        }[self.direction]
        return {
            None: shorthand,
            True: f"{shorthand} expct",
            False: f"{shorthand} corr"
        }[self.expected]

    @property
    def column(self):
        return {
            None: self.direction,
            True: f"{self.direction}_expected",
            False: f"{self.direction}_correction"
        }[self.expected]

    @property
    def error_column(self):
        return f"error_{self.column}"


# Plotting ---------------------------------------------------------------------

def plot_segment_data(
    widget: DualPlotWidget, 
    definitions: DualPlotDefinition, 
    segments: list[SegmentDataModel], 
    settings: PlotSettings
    ):
    """ Plot the given segments with the given definition. """
    # use the segment name as label, if there is more than one segment from the same measurement
    use_segment_label = len(set(s.measurement.display() for s in segments)) != len(segments)
    def get_label(segment: SegmentDataModel) -> str:
        if use_segment_label:
            return f"{segment.measurement.display()} {segment.name}"
        return segment.measurement.display()

    # wrap loading for better error handling and logging
    def get_data(segment: SegmentDataModel, file_name: str):
        try:
            return segment.data[file_name]
        except FileNotFoundError:
            LOGGER.error(f"Segment {segment.name} has no data for {file_name}.")
            return None
    
    # set x column 
    x_column = S_COLUMN
    if settings.model_s:
        x_column = S_MODEL_COLUMN

    # Loop over top/bottom plots ---
    for definition, plot in zip(definitions.plots, widget.plots):
        definition: PlotDefinition

        dataframes = {
            get_label(segment): get_data(segment, definition.file_name) 
            for segment in segments
        }

        if any(df is None for df in dataframes.values()):
            LOGGER.error("Could not find data for all segments, please run these again !?")
            # continue anyway
            dataframes = {label: df for label, df in dataframes.items() if df is not None}

        # Plot Model Elements ---
        if settings.show_model:
            model_dir = segments[0].measurement.model_dir
            bpm_ranges = [(s.start, s.end) for s in segments]
            plot_element_lines(
                plot=plot,
                data_frame=load_twiss_elements(model_dir),
                ranges=bpm_ranges,
                start_zero=not settings.model_s,
            )

        # Loop over forward/backward plots ---
        for direction in ("forward", "backward"):
            if not getattr(settings, direction):  # user activated
                continue
            
            # Loop over propagaed/expected or corrected values --- 
            for expected in (None, settings.expected):
                style = DirectionStyle(direction, expected)

                # Now put it all together ---
                plot_dataframes(
                    plot=plot, 
                    dataframes=dataframes, 
                    xcolumn=x_column.column, 
                    ycolumn=getattr(definition.propagable_columns, style.column),
                    yerrcolumn=getattr(definition.propagable_columns, style.error_column),
                    xlabel=x_column.label,
                    ylabel=definition.ylabel,
                    legend=settings.show_legend,
                    marker=style.marker,
                    markersize=settings.marker_size,
                    brightness=style.brightness,
                    linestyle=style.linestyle,
                    suffix=style.suffix,
                )

        if settings.reset_zoom:
            plot.enableAutoRange()


@cache
def load_twiss_elements(model_dir: Path) -> tfs.TfsDataFrame:
    """ Load the twiss elements from the model directory. 
    Cache here, because that might take a moment, so better to keep the DataFrame in memory.
    """
    df = tfs.read_tfs(model_dir / TWISS_ELEMENTS_DAT, index=NAME)
    df = df.loc[df.index.str.match(r"^(?!DRIFT).*"), [S_COLUMN.column]]  # we actually only need the s column and headers
    return df



""" 
Latex to HTML converter
-----------------------

Converts LaTeX commands for Greek letters and other symbols to HTML.
Needed to be able to re-use matplotlib-labels - which understand Latex - 
in pyqtgraph, which does not understand Latex.
"""
from __future__ import annotations

import re

# Dictionary to map Greek LaTeX symbols to HTML
LATEX_TO_HTML_SYMBOLS = {
    # Greek ---
    r'\alpha': 'α',
    r'\beta': 'β',
    r'\gamma': 'γ',
    r'\delta': 'δ',
    r'\epsilon': 'ε',
    r'\zeta': 'ζ',
    r'\eta': 'η',
    r'\theta': 'θ',
    r'\iota': 'ι',
    r'\kappa': 'κ',
    r'\lambda': 'λ',
    r'\mu': 'μ',
    r'\nu': 'ν',
    r'\xi': 'ξ',
    r'\pi': 'π',
    r'\rho': 'ρ',
    r'\sigma': 'σ',
    r'\tau': 'τ',
    r'\upsilon': 'υ',
    r'\phi': 'φ',
    r'\chi': 'χ',
    r'\psi': 'ψ',
    r'\omega': 'ω',
    r'\Alpha': 'Α',
    r'\Beta': 'Β',
    r'\Gamma': 'Γ',
    r'\Delta': 'Δ',
    r'\Epsilon': 'Ε',
    r'\Zeta': 'Ζ',
    r'\Eta': 'Η',
    r'\Theta': 'Θ',
    r'\Iota': 'Ι',
    r'\Kappa': 'Κ',
    r'\Lambda': 'Λ',
    r'\Mu': 'Μ',
    r'\Nu': 'Ν',
    r'\Xi': 'Ξ',
    r'\Pi': 'Π',
    r'\Rho': 'Ρ',
    r'\Sigma': 'Σ',
    r'\Tau': 'Τ',
    r'\Upsilon': 'Υ',
    r'\Phi': 'Φ',
    r'\Chi': 'Χ',
    r'\Psi': 'Ψ',
    r'\Omega': 'Ω',
    # Symbols ---
    r'\pm': '&plusmn;',
    r'\times': '&times;',
    r'\Re': '&real;',
    r'\Im': '&image;',
    # Spacing ---
    r'\quad': '&ensp;',
    r'\;': '&thinsp;',
}

def latex_to_html_converter(latex_str: str) -> str:
    """ 
    Converts LaTeX commands for Greek letters and other symbols to HTML.

    Args:
        latex_str (str): LaTeX string to convert.
    
    Returns:
        str: HTML string
    """
    # Convert LaTeX commands for Greek letters to HTML
    html_str = latex_str.replace("$", "")
    for latex, html in LATEX_TO_HTML_SYMBOLS.items():
        html_str = html_str.replace(latex, html)
    
    # Other HTML formatting like superscript/subscript, fractions, etc.
    html_str = re.sub(r'_{([^}]*)}', r'<sub>\1</sub>', html_str)
    html_str = re.sub(r'_(.)', r'<sub>\1</sub>', html_str)
    html_str = re.sub(r'\\frac{([^}]*)}{([^}]*)}', r'<sup>\1</sup>/<sub>\2</sub>', html_str)
    html_str = re.sub(r'\\left\((.*?)\\right\)', r'(\1)', html_str)  # Basic parentheses
    html_str = re.sub(r'\\left\|(.*?)\\right\|', r'|\1|', html_str)  # Absolute values
    
    return html_str



""" 
Iterable Classes
----------------

This module contains classes that are or make other classes iterable over their attributes.
"""
from typing import Any
from collections.abc import Iterator

EXCLUDED_NAME = "EXCLUDED_ATTRIBUTES"

# Metaclasses ------------------------------------------------------------------

class IterableAttributeNames(type):
    """ Makes the class itself iterable over its attribute names. """

    def __iter__(self) -> Iterator[str]:
        for attr in self.__dict__.keys():
            if not attr.startswith("__") and attr != EXCLUDED_NAME and attr not in getattr(self, EXCLUDED_NAME, []):
                yield attr


class IterableAttributeValues(type):
    """ Makes the class itself iterable over its attribute values. """
    def __iter__(self) -> Iterator[Any]:
        for attr, value in self.__dict__.items():
            if not attr.startswith("__"):
                yield value


class IterableAttributeItems(type):
    """ Makes the class itself iterable over its attribute name and values. """
    def __iter__(self) -> Iterator[tuple[str, Any]]:
        for attr, value in self.__dict__.items():
            if not attr.startswith("__"):
                yield attr, value


# Iterable Class ---------------------------------------------------------------


class IterClass(metaclass=IterableAttributeNames):

    EXCLUDED_ATTRIBUTES = ["keys", "values", "items"]
    
    @classmethod
    def keys(cls) -> Iterator[str]:
        for attr in cls:
            yield attr
    
    @classmethod
    def values(cls) -> Iterator[Any]:
        for attr in cls:
            yield getattr(cls, attr)
    
    @classmethod
    def items(cls) -> Iterator[tuple[str, Any]]:
        for attr in cls:
            yield attr, getattr(cls, attr)



""" 
DataClass UI
------------

A simple UI for dataclasses, which allows to edit the values of dataclasses
by the user within a dialog.
"""
from omc3_gui.ui_components.dataclass_ui import controller, view, model

# Dialogs ---
DataClassDialog = view.DataClassDialog
SettingsDialog = view.SettingsDialog

# Controller ---
DataClassUI = controller.DataClassUI
DataClassTabbedUI = controller.DataClassTabbedUI

# Model ---
FieldUIDef = model.FieldUIDef
DirectoryPath = model.DirectoryPath
FilePath = model.FilePath

metafield = model.metafield
choices_validator = model.choices_validator


""" 
TFS Plotter
-----------

This module contains functions to plot TFS dataframes with pyqtgraph.
"""
from collections.abc import Sequence

import numpy as np
import pandas as pd
import pyqtgraph as pg
from omc3.plotting.utils.colors import get_mpl_color
import logging

from qtpy.QtCore import Qt
from qtpy.QtGui import QColor

from omc3_gui.plotting.classes import ObservablePlotDataItem, YAXES_WIDTH

PenStyle = Qt.PenStyle

LOGGER = logging.getLogger(__name__)


def plot_dataframes(
    plot: pg.PlotWidget, 
    dataframes: dict[str, pd.DataFrame], 
    xcolumn: str, 
    ycolumn: str, 
    xerrcolumn: str = None,
    yerrcolumn: str = None,
    xlabel: str = None, 
    ylabel: str = None,
    legend: bool = True,
    brightness: int | None = None,
    marker: str = 'o',
    markersize: float = 6,
    linestyle: PenStyle = PenStyle.SolidLine,
    suffix: str = "",
    ):
    """ 
    Plot a collection of DataFrames with pyqtgraph.

    Args:
        plot (pg.PlotWidget): The plot to plot the dataframes into.
        dataframes (dict[str, pd.DataFrame]): A dictionary of DataFrames to plot.
        xcolumn (str): The name of the column to plot on the x-axis.
        ycolumn (str): The name of the column to plot on the y-axis.
        xerrcolumn (str): The name of the column to plot as horizontal errorbars.
        yerrcolumn (str): The name of the column to plot as vertical errorbars.
        xlabel (str): The label of the x-axis.
        ylabel (str): The label of the y-axis.
        legend (bool, optional): Whether to add a legend to the plot. Defaults to True.
        brightness (int, optional): The brightness of the colors to use. Defaults to None.
        marker (str, optional): The marker to use for the data points. Defaults to 'o'.
        markersize (float, optional): The size of the markers to use for the data points. Defaults to 6.
        linestyle (PenStyle, optional): The linestyle to use for the data points. Defaults to PenStyle.SolidLine.
        suffix (str, optional): The suffix to add to the legend. Defaults to "".
    """
    plot_item: pg.PlotItem = plot.plotItem
    
    plot_item.addLegend(offset=(0, 0))
    
    for idx, (name, df) in enumerate(dataframes.items()):
        color = pg.Color(get_mpl_color(idx))
        if brightness is not None:
            color = color.lighter(brightness)
        
        try:
            df[ycolumn]
        except KeyError:
            LOGGER.debug(f"Could not find column '{ycolumn}' in DataFrame for '{name}'. Skipping!")
            continue

        plot_errorbar(
            plot_item, 
            x=df[xcolumn], 
            y=df[ycolumn], 
            xerr=df.get(xerrcolumn), 
            yerr=df.get(yerrcolumn), 
            names=df.index, 
            label=f"{name}{suffix}", 
            color=color,
            marker=marker,
            markersize=markersize,
            linestyle=linestyle,
        )
    
    if xlabel is not None:
        plot_item.setLabel("bottom", xlabel)

    if ylabel is not None:
        plot_item.setLabel("left", ylabel)
        
    plot_item.getAxis("left").setWidth(YAXES_WIDTH)  # keep constant
    plot_item.legend.setVisible(legend)


def plot_errorbar(
    plot: pg.PlotItem,
    *,
    x: Sequence, 
    y: Sequence, 
    xerr: Sequence | None = None, 
    yerr: Sequence | None = None, 
    names: Sequence | None = None, 
    label: str | None = None,
    color: str | pg.Color | None = None,
    marker: str = 'o', 
    markersize: float = 6,
    linestyle: PenStyle = PenStyle.SolidLine,
    linewidth: float = 2,
    ) -> tuple[pg.PlotDataItem, pg.ErrorBarItem]:
    """ 
    Plot a single errorbar with pyqtgraph.
    This tries to imitate the behavior of matplotlib's errorbar function, 
    and the naming is mostly borrowed from there.

    Args:
        plot (pg.PlotWidget): The plot to plot the errorbar into.
        x (Sequence): The x values of the errorbar.
        y (Sequence): The y values of the errorbar.
        xerr (Sequence): The xerr values of the errorbar.
        yerr (Sequence): The yerr values of the errorbar.
        names (Sequence): The names of the entries in the data sequence.
        label (str | None, optional): The label of the errorbar. Defaults to None.
        color (str | None, optional): The color of the errorbar. Defaults to None.
        marker (str, optional): The marker of the errorbar. Defaults to 'o'.
        markersize (float, optional): The markersize of the errorbar. Defaults to 10.
        linestyle (PenStyle, optional): The linestyle of the errorbar. Defaults to PenStyle.SolidLine.
        linewidth (float, optional): The linewidth of the errorbar. Defaults to 2.
    """

    curvePen = pg.mkPen(color=color, width=linewidth, style=linestyle)    
    errorbarPen = pg.mkPen(color=color, width=linewidth, style=PenStyle.SolidLine)
    
    # convert everything to numpy, as this is what pyqtgraph expects. 
    # pd.Series seems to also work for now, but raises deprecation warnings, 
    # as in the future it needs .iloc to work with indices, 
    # yet pyqtgraph and create_tooltips access items via `[ ]` (for now at least).
    x = safe_convert_to_numpy(x)
    y = safe_convert_to_numpy(y)
    xerr = safe_convert_to_numpy(xerr)
    yerr = safe_convert_to_numpy(yerr)
    names = safe_convert_to_numpy(names)

    hex_color = None
    if color is not None:
        hex_color = curvePen.color().name(QColor.NameFormat.HexRgb)
    
    tooltips = create_tooltips(x, y, xerr, yerr, names, label, hex_color)
    curve = ObservablePlotDataItem(
        x=x, y=y, data=tooltips,
        name=label,
        pen=curvePen, 
        tip=None,
        symbol=marker, symbolBrush=color, symbolSize=markersize
    )
    # curve.sigPointsHovered.connect(hovered)
    # curve.sigPointsClicked.connect(clicked)
    curve.scatter.opts['hoverable'] = True
    curve.scatter.sigHovered.connect(hovered)
    
    errorbar = None
    if xerr is not None or yerr is not None:
        errorbar = pg.ErrorBarItem(
            x=x, y=y, data=tooltips, 
            width=2*xerr if xerr is not None else None, 
            height=2*yerr if yerr is not None else None, 
            pen=errorbarPen
        )
        
        # Connect errorbar and curve's visibility
        curve.visibilityChanged.connect(errorbar.setVisible)
        plot.addItem(errorbar)

    plot.addItem(curve)
    return curve, errorbar


def create_tooltips(x, y, xerr, yerr, names, label, color) -> list[str]:
    """
    Create a list of tooltips for a given errorbar.

    Args:
        x (Sequence): The x values of the errorbar.
        y (Sequence): The y values of the errorbar.
        xerr (Sequence): The xerr values of the errorbar.
        yerr (Sequence): The yerr values of the errorbar.
        names (Sequence): The names of the entries in the data sequence.
        label (str | None, optional): The label of the errorbar
        color (str | None, optional): The color of the tooltip background
    """
    tooltips = [""] * len(x)

     
    for index in range(len(x)):
        tooltip_text = "<html>"

        if color is not None:
            tooltip_text +=  "" # TODO

        if label is not None:
            tooltip_text += f"{label}<br>"

        tooltip_text += f"x: {x[index]:.2e}"
        if xerr is not None:
            tooltip_text += f" ± {xerr[index]:.2e}"
        
        tooltip_text += f"<br>y: {y[index]:.2e}"
        if yerr is not None:
            tooltip_text += f" ± {yerr[index]:.2e}"
        
        if names is not None:
            tooltip_text += f"<br>{names[index]}"
        
        if color is not None:
            tooltip_text += ""
            
        tooltip_text += "</html>"

        tooltips[index] = tooltip_text
    return tooltips


def safe_convert_to_numpy(data: Sequence | None) -> list:
    if data is None:
        return None
    
    # check if data is even iterable
    try:
        return data.to_numpy()
    except AttributeError:
        return np.array(data)

def hovered(item, points, ev):
    if not len(points):
        item.setToolTip(None)
        return
    item.setToolTip(points[0].data())


def clicked(item, points, ev):
    # print('clicked')
    pass




"""
UI: File Dialogs
----------------

Helper functions to open files.
"""
import logging
from pathlib import Path

from qtpy.QtWidgets import QApplication, QFileDialog, QStyle

LOGGER = logging.getLogger(__name__)


# Open Dialog Windows ----------------------------------------------------------
class OpenFilesDialog(QFileDialog):
    """ Quick dialog to open any kind of file. 
    Modifies QFileDialog, and allows only kwargs to be passed. 
    """
    
    def __init__(self, **kwargs) -> None:
        if "directory" in kwargs and isinstance(kwargs["directory"], Path): 
            kwargs["directory"] = str(kwargs["directory"])  # allow giving Paths

        super().__init__(**kwargs)  # parent, caption, directory, filter, options
        self.setOption(QFileDialog.Option.DontUseNativeDialog, True)

    def run_selection_dialog(self) -> list[Path]:
      if self.exec_():
         return [Path(f) for f in self.selectedFiles()]
      return []


class OpenFileDialog(OpenFilesDialog):
    """ Open a single file. """

    def __init__(self, caption: str = "Select File", **kwargs) -> None:
        super().__init__(caption=caption, **kwargs)  # parent, directory, filter, options
        self.setFileMode(QFileDialog.FileMode.ExistingFile)
    
    def run_selection_dialog(self) -> Path:
        selected = super().run_selection_dialog()
        if selected:
            return selected[0]
        return None


class OpenDirectoriesDialog(OpenFilesDialog):
    """ Open multiple directories. """

    def __init__(self, caption: str = "Select Folders", **kwargs) -> None:
        super().__init__(caption=caption, **kwargs)  # parent, directory, filter, options
        icon = QApplication.style().standardIcon(QStyle.StandardPixmap.SP_DirIcon)
        self.setWindowIcon(icon)
        self.setOption(QFileDialog.Option.ShowDirsOnly, True)
        self.setFileMode(QFileDialog.FileMode.ExistingFiles) 

    def accept(self):
        """This function is called when the user clicks on "Open".
        Normally, when selecting a directories, the first directory is followed/opened inside the dialog, 
        i.e. its content is shown. Overwrite super().accept() to prevent that and close the dialog instead.
        """
        for selected in self.selectedFiles():
            if not Path(selected).is_dir():  # this should not happen, as only directories are shown
                LOGGER.warning(f"{selected} is not a directory. Try again.")
                return

        self.done(QFileDialog.Accepted)


class OpenDirectoryDialog(OpenDirectoriesDialog):
    """ Open a single directory. """

    def __init__(self, caption: str = "Select Folder", **kwargs) -> None:
        super().__init__(caption=caption, **kwargs)  # parent, directory, filter, options
        self.setFileMode(QFileDialog.FileMode.Directory)
    
    def run_selection_dialog(self) -> Path:
        selected = super().run_selection_dialog()
        if selected:
            return selected[0]
        return None


class OpenAnyMultiDialog(OpenFilesDialog):
    """ Open multiple files/folders. """

    def __init__(self, caption: str = "Select Files", existing: bool = True, **kwargs) -> None:
        super().__init__(caption=caption, **kwargs)  # parent, directory, filter, options
        if existing:
            self.setFileMode(QFileDialog.FileMode.ExistingFiles)
    
    def accept(self):
        """This function is called when the user clicks on "Open".
        Normally, when selecting a directories, the first directory is followed/opened inside the dialog, 
        i.e. its content is shown. 
        Overwrite super().accept() to prevent that and close the dialog instead.
        """
        if not self.selectedFiles():
            LOGGER.warning("Nothing selected. Try again or cancel.")
            return

        self.done(QFileDialog.Accepted)
    

class OpenAnySingleDialog(OpenAnyMultiDialog):
    """ Open a single file/folder. """

    def __init__(self, caption: str = "Select Single", **kwargs) -> None:
        existing = kwargs.get("existing", True)
        super().__init__(caption=caption, **kwargs)  # parent, directory, filter, options
        if existing:
            self.setFileMode(QFileDialog.FileMode.ExistingFile)
    
    def run_selection_dialog(self) -> Path:
        selected = super().run_selection_dialog()
        if selected:
            return selected[0]
        return None



""" 
UI: Message Boxes
-----------------

Helper functions to display message boxes.
"""
from qtpy.QtWidgets import QMessageBox, QWidget

def show_confirmation_dialog(question: str, title: str = "Confirmation", parent: QWidget = None) -> bool:
    """ Displays a confirmation dialog.

    Could also be done with QMessageBox.question(parent, title, question, QMessageBox.Ok | QMessageBox.Cancel).

    Args:
        question (str): Dialog question.
        title (str): Dialog title.
        parent (QtWidgets.QWidget): Parent widget.
    
    Returns:
        bool: True if the user confirmed, False otherwise
    """
    msg_box = QMessageBox(
        QMessageBox.Question,
        title,
        question,
        QMessageBox.Ok | QMessageBox.Cancel,
        parent
    )
    msg_box.setDefaultButton(QMessageBox.Cancel)
    result = msg_box.exec_()
    return result == QMessageBox.Ok


def show_error_dialog(message: str, title: str = "Error", parent: QWidget = None):
    """ Displays an error dialog. 
    
    This is a convenience function to displays an error dialog.

    Args:
        title (str): Dialog title.
        message (str): Dialog message.
        parent (QtWidgets.QWidget): Parent widget.
    """
    message_box = QMessageBox(
        QMessageBox.Critical,
        title,
        message,
        QMessageBox.Ok,
        parent,
    )
    message_box.exec_()


"""
UI: Widgets 
-----------

Pre-Defined Widgets go here.
"""
from __future__ import annotations

import math
from qtpy import QtWidgets, QtCore, QtGui
from omc3_gui.ui_components import colors


# Buttons ----------------------------------------------------------------------

class RunButton(QtWidgets.QPushButton):
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not args and "text" not in kwargs:
            self.setText("Run")

        self.setStyleSheet(
            f":enabled {{ background-color: {colors.GREEN_DARK}; color: {colors.TEXT_LIGHT}; }}"
            f":disabled {{ background-color: {colors.GREEN_DARK_GREY}; color: {colors.GREYED_OUT_TEXT_DARK}; }}"
        )


class OpenButton(QtWidgets.QPushButton):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not args and "text" not in kwargs:
            self.setText("Open")

        self.setStyleSheet(
            f":enabled {{ background-color: {colors.GREEN_LIGHT}; color: {colors.TEXT_DARK}; }}"
            f":disabled {{ background-color: {colors.GREEN_LIGHT_GREY}; color: {colors.GREYED_OUT_TEXT_DARK}; }}"
            )


class RemoveButton(QtWidgets.QPushButton):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not args and "text" not in kwargs:
            self.setText("Remove")

        self.setStyleSheet(
            f":enabled {{ background-color: {colors.RED_DARK}; color: {colors.TEXT_LIGHT}; }}"
            f":disabled {{ background-color: {colors.RED_GREY}; color: {colors.GREYED_OUT_TEXT_DARK}; }}"
            )

class ChangeButton(QtWidgets.QPushButton):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setStyleSheet(
            f":enabled {{ background-color: {colors.BLUE_DARK}; color: {colors.TEXT_LIGHT}; }}"
            f":disabled {{ background-color: {colors.BLUE_GREY}; color: {colors.GREYED_OUT_TEXT_DARK}; }}"
            )

class DefaultButton(QtWidgets.QPushButton):
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

# Filler -----------------------------------------------------------------------

class HorizontalSeparator(QtWidgets.QFrame):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setFrameShape(QtWidgets.QFrame.HLine)
        self.setFrameShadow(QtWidgets.QFrame.Sunken) 


class VerticalSeparator(QtWidgets.QFrame):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setFrameShape(QtWidgets.QFrame.VLine)
        self.setFrameShadow(QtWidgets.QFrame.Sunken)

# Animations -------------------------------------------------------------------

class RunningSpinner(QtWidgets.QWidget):

    def __init__(self, center_on_parent=True, *args, **kwargs):
        self._spinner_color: QtGui.QColor = kwargs.pop("color", QtGui.QColor(colors.BLUE_DARK))
        self._spinner_roundness: float = kwargs.pop("roundness", 300.0)
        self._spinner_min_trail_opacity: float = kwargs.pop("min_trail_opacity", 0.05)
        self._spinner_trail_fade_percentage: float = kwargs.pop("trail_fade_percentage", 100.0)
        self._spinner_revolutions_per_second: float = kwargs.pop("revolutions_per_second", 1.2)
        self._spinner_n_lines: int = kwargs.pop("n_lines", 10)
        self._spinner_length: float = kwargs.pop("length", 5)
        self._spinner_radius: float = kwargs.pop("radius", 3)
        self._spinner_rotate: float = kwargs.pop("rotate", False)
        self._spinner_width: float = self._spinner_radius * 2 * math.pi / self._spinner_n_lines
        self._spinner_center: bool = center_on_parent

        self._angle: int = 0
        self._is_spinning: bool = False

        QtWidgets.QWidget.__init__(self, *args, **kwargs)
        
        self._animation = QtCore.QPropertyAnimation(self, b"angle", self)
        self._animation.setStartValue(0)
        self._animation.setEndValue(360)
        self._animation.setLoopCount(-1)
        self._animation.setDuration(int(1000 / self._spinner_revolutions_per_second))

        self.updateSize()

    @QtCore.Property(int)
    def angle(self):
        return self._angle

    @angle.setter
    def angle(self, value):
        self._angle = value % 360
        self.update()

    def updateSize(self):
        size = (self._spinner_radius + self._spinner_length) * 2
        self.setFixedSize(size, size)

    def updatePosition(self):
        parent = self.parentWidget()
        if parent and self._spinner_center:
            self.move(int(parent.width() / 2 - self.width() / 2),
                      int(parent.height() / 2 - self.height() / 2))

    def currentLineColor(self, line_number):
        color = QtGui.QColor(self._spinner_color)
        if line_number == 0:
            return color

        distance_threshold = math.ceil((self._spinner_n_lines - 1) * self._spinner_trail_fade_percentage / 100.0)
        if line_number > distance_threshold:
            color.setAlphaF(self._spinner_min_trail_opacity)

        else:
            alpha_diff = 1 - self._spinner_min_trail_opacity
            gradient = alpha_diff / distance_threshold
            alpha = color.alphaF() - gradient * line_number 
            alpha = min(1.0, max(0.0, alpha))
            color.setAlphaF(alpha)
        return color

    def paintEvent(self, event):
        self.updatePosition()
        painter = QtGui.QPainter(self)
        painter.fillRect(self.rect(), QtCore.Qt.transparent)
        painter.setRenderHint(QtGui.QPainter.Antialiasing, True)
        painter.setPen(QtCore.Qt.NoPen)

        for idx_line in range(self._spinner_n_lines):
            painter.save()
            painter.translate(self._spinner_radius + self._spinner_length,
                              self._spinner_radius + self._spinner_length)
            rotateAngle = -(360.0 * idx_line / self._spinner_n_lines) + (self.angle if self._spinner_rotate else 0)
            color_index = idx_line + (int(self.angle / 360.0 * self._spinner_n_lines) if not self._spinner_rotate else 0)
            painter.rotate(rotateAngle)
            painter.translate(self._spinner_radius, 0)
            color = self.currentLineColor(color_index % self._spinner_n_lines)
            painter.setBrush(color)
            
            painter.drawRoundedRect(
                QtCore.QRect(int(0), -int(self._spinner_width / 2), int(self._spinner_length), int(self._spinner_width)),
                int(self._spinner_roundness),
                int(self._spinner_roundness), 
                QtCore.Qt.RelativeSize
            )
            painter.restore()

    def start(self):
        self.updatePosition()
        self._is_spinning = True
        self.show()
        self._animation.start()

    def stop(self):
        self._animation.stop()
        self._is_spinning = False
        self.hide()

    @property
    def is_spinning(self):
        return self._is_spinning



""" 
Counter
-------

Classes to do some automatic counting and filling. 
"""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from qtpy import QtWidgets


class Counter:
    """ Simple class to count up integers. Similar to itertools.count """

    def __init__(self, start=0, end=None):
        self.count = start
        self._end = end
        self._start = start

    def reset(self):
        self.count = self._start

    def __next__(self):
        self.count += 1

        if self._end is not None and self.count >= self._end:
            raise StopIteration

        return self.count

    def __iter__(self):
        return self

    def next(self):
        return next(self)

    def current(self):
        return self.count


class HorizontalGridLayoutFiller:
    """
    Fills a grid-layout with widgets, without having to give row and col positions, 
    but allows giving a col-span.
    """

    def __init__(self, layout: QtWidgets.QGridLayout, cols: int, rows: int = None):
        self._layout = layout
        self._cols = cols
        self._rows = rows
        self._current_col = 0
        self._current_row = 0

        
    def add(self, widget, col_span=1):
        self._layout.addWidget(widget, self._current_row, self._current_col, 1, col_span)
        self._current_col += col_span
        if self._current_col > self._cols:
            raise ValueError("Span too large for given columns.")

        if self._current_col == self._cols:
            self._current_col = 0
            self._current_row += 1
            if self._rows is not None and self._current_row >= self._rows:
                raise ValueError("Grid is already full.")

    addWidget = add

        


""" 
UI: Base Classes for CVM
------------------------

This module contains base classes for UI's
that use the Controller-View-Model pattern.
"""
from __future__ import annotations

import logging
import re
import sys

from qtpy import QtGui
from qtpy.QtCore import QObject, Slot
from qtpy.QtWidgets import (
    QAction,
    QApplication,
    QDesktopWidget,
    QDockWidget,
    QMenu,
    QMenuBar,
    QStatusBar,
    QStyle,
    QWidgetAction,
)

from omc3_gui import __version__
from omc3_gui.ui_components import colors
from omc3_gui.utils.log_handler import get_console_formatter
from omc3_gui.ui_components.widgets import RunningSpinner
from omc3_gui.ui_components.message_boxes import show_error_dialog

try:  # CERN Application Frame
    from accwidgets.app_frame import ApplicationFrame
    from accwidgets.qt import exec_app_interruptable
except ImportError:  # Standard QT
    from qtpy.QtWidgets import QMainWindow as ApplicationFrame

    def exec_app_interruptable(app):
        app.exec_()

try:  # CERN Console
    from accwidgets.app_frame._about_dialog import AboutDialog
    from accwidgets.log_console import LogConsoleFormatter as AccPyLogConsoleFormatter
except ImportError:  # Deactivated
    AccPyLogConsoleFormatter = object
    AboutDialog = None

LOGGER = logging.getLogger(__name__)


class Controller(QObject):
    """ 
    Base class for the controller of a UI.

    The controller is the glue between the view and the model
    and also the entry-point for the app.
    """

    def __init__(self, view: ApplicationFrame, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._view = view

    def show(self):
        self._view.show()

    @classmethod
    def run_application(cls, *args, **kwargs) -> int:
        app = QApplication(sys.argv)
        controller = cls(*args, **kwargs)
        controller.show()
        return exec_app_interruptable(app)
    

class View(ApplicationFrame):
    """ 
    Base class for the view of a UI.

    Adds a menu bar and a status bar as well as the log-console (if in CERN mode).
    """
    def __init__(self, *args, **kwargs):
        kwargs["use_log_console"] = kwargs.get("use_log_console", True)
        try:
            super().__init__(*args, **kwargs)  # CERN Application Frame
        except TypeError:
            del kwargs["use_log_console"]
            super().__init__(*args, **kwargs)   # QT Main window
        
        self._menu_bar: QMenuBar = None
        self._thread_spinner: RunningSpinner = None
        self.app_version = __version__

        self._adapt_logger()
        self._adapt_to_screensize()
        self.build_menu_bar()
        self.build_status_bar()

    def build_menu_bar(self):
        self._menu_bar = QMenuBar()
        
        # File menu ---
        file = self._menu_bar.addMenu("File")
        quit = file.addAction("Exit", self.close)
        quit.setIcon(QApplication.style().standardIcon(QStyle.StandardPixmap.SP_DialogCloseButton))
        quit.setMenuRole(QWidgetAction.QuitRole)

        # View menu ---
        # needs to be called "View" to be found by ApplicationFrame,
        # which adds some additional actions if needed.
        view = self._menu_bar.addMenu("View") 

        # Fullscreen -
        toggle_fullscreen = view.addAction("Full Screen", self.toggleFullScreen)
        toggle_fullscreen.setCheckable(True)

        # Log Console -
        # Add the log-console checkbox here, so Application Frame doesn't give it 
        # the wrong title ("Toggle Log Console"), is horrible for a checkbox (jdilly, 2025)
        log_console: QDockWidget = getattr(self, "log_console", None)
        if log_console:
            log_console_action = log_console.toggleViewAction()
            log_console_action.setText("Log Console")
            # hide the setText function, as otherwise ApplicationFrame overwrites the title
            log_console_action._setText = log_console_action.setText
            log_console_action.setText = lambda text: None  
            view.addAction(log_console_action) 

        # Help menu ---
        help = self._menu_bar.addMenu("Help")

        # About -
        about = help.addAction("About", self.showAboutDialog)
        about.setIcon(self.windowIcon())
        about.setMenuRole(QWidgetAction.AboutRole)

        # Set menu bar ---
        self.setMenuBar(self._menu_bar)
    
    def get_action_by_title(self, title: str, parent: QMenuBar | QMenu | None = None) -> QAction | QMenu:
        """ Retrieve a menu action by its title. 
        
        Args:
            title (str): Action title.
            parent (QMenuBar): Parent menu bar to search, if `None` the main menu bar is used.
        """
        if parent is None:
            parent: QMenuBar | None = self._menu_bar

        if parent is None:
            LOGGER.debug("Menu bar does not seem to have been build yet.")
            return None

        for action in parent.actions():
            if action.text() == title:
                menu = action.menu()
                if menu is not None:  # action is a menu (submenu)
                    return menu
                return action  # action is an entry (leaf)

        LOGGER.debug(f"Unable to find action with title: {title} in {parent!r}")
        return None

    def build_status_bar(self):
        status_bar: QStatusBar = self.statusBar()
        
        # Build thread spinner ---
        thread_spinner = RunningSpinner(parent=status_bar, center_on_parent=False)
        status_bar.addPermanentWidget(thread_spinner)
        thread_spinner.mouseDoubleClickEvent = lambda *args, **kwargs: self.sig_thread_spinner_double_clicked.emit()
        thread_spinner.stop()
        self._thread_spinner = thread_spinner
       
        self.setStatusBar(status_bar)
        # status_bar.hide()  # looks nice, but moves the window around too much ... 
    
    @property
    def thread_spinner(self) -> RunningSpinner:
        return self._thread_spinner

    def _adapt_to_screensize(self):
        """ Sets the window size to 2/3 of the screen size. """
        screen_shape = QDesktopWidget().screenGeometry()
        self.resize(
            int(2 * screen_shape.width() / 3),
            int(2 * screen_shape.height() / 3)
        )

    def _adapt_logger(self):
        """ Changes the appearance of the log console. """
        if getattr(self, "log_console") is None:
            return
        self.log_console.console.expanded = False
        self.log_console.setFeatures(
            self.log_console.DockWidgetClosable | self.log_console.DockWidgetMovable
        )
        self.log_console.console.formatter = LogConsoleFormatter(show_date=False)  # see below
        for level, color in colors.LOGGING.items():
            self.log_console.console._set_color_to_scheme(color=QtGui.QColor(color), level=level)
        self.log_console.console.model.buffer_size = 10_000  # default: 1000
        if sys.flags.debug:
            self.log_console.console.model.visible_levels |=  {logging.DEBUG}

    def setWindowTitle(self, title: str):
        super().setWindowTitle(f"{title} v{self.app_version}")

    def toggleFullScreen(self):
        if self.isFullScreen():
            self.showNormal()
        else:
            self.showFullScreen()

    @Slot()
    def showAboutDialog(self) -> None:
        """
        Display an 'about' dialog.
        """
        if AboutDialog is None:
            return

        name = self.windowTitle()
        try:
            name = " ".join(name.split()[:-1])
        except (IndexError, TypeError, AttributeError):
            pass

        dialog = AboutDialog(
            app_name=name,
            version=self.app_version,
            icon=self.windowIcon(),
            parent=self,
        )
        dialog.exec_()

    def showErrorDialog(self, title: str, message: str):
        """ Convenience function to displays an error dialog. 
        
        Args:
            title (str): Dialog title.
            message (str): Dialog message.
        """
        LOGGER.error(message)
        show_error_dialog(message, title, parent=self)


class LogConsoleFormatter(AccPyLogConsoleFormatter):

    def __init__(self, show_date: bool = True, show_time: bool = True, show_logger_name: bool = True) -> None:
        """
        Reimplementation of the AccPy LogConsoleFormatter, to allow for a different logging format.

        Args:
            show_date: Add date to the log message prefix.
            show_time: Add time to the log message prefix.
            show_logger_name: Add logger name to the log message prefix.
        """
        super().__init__()

        self.show_date = show_date
        self.show_time = show_time
        self.show_logger_name = show_logger_name

        fmt_str = get_console_formatter()._fmt
        date_format = []

        if show_date:
            date_format.append("%Y-%m-%d")
        if show_time:
            date_format.append("%H:%M:%S")
        
        if not show_date and not show_time:
            fmt_str = re.sub(r"\%\(asctime\)\d+s\s+", "", fmt_str)

        if not show_logger_name:
            fmt_str = re.sub(r"\%\(name\)\d+s\s+", "", fmt_str)
        
        fmt_str = fmt_str[fmt_str.index("%"):]

        self._fmt = logging.Formatter(fmt=fmt_str, datefmt=" ".join(date_format), style="%")



""" 
Logging
-------

Helper functions for logging.
"""
import sys
import os
import logging
from logging import StreamHandler, FileHandler


def set_up_console_logger(logger):
    logger.setLevel(logging.DEBUG)

    console_handler = StreamHandler(sys.stdout)
    console_handler.setLevel(logging.INFO)

    logger.addHandler(console_handler)
    return logger


def add_file_handler(log_dir):
    logger = logging.getLogger("")
    log_file = os.path.join(log_dir, "log.txt")
    file_handler = FileHandler(log_file, mode="w")
    file_handler.setLevel(logging.DEBUG)
    file_handler.setFormatter(get_file_formatter())
    logger.addHandler(file_handler)
    logger.info("Set up debug log file in: " +
                os.path.abspath(log_file))


def get_file_formatter():
    formatter = logging.Formatter(
        "%(name)s %(asctime)s %(levelname)s %(message)s"
    )
    formatter.datefmt = '%Y-%m-%d %H:%M:%S'
    return formatter


def get_console_formatter():
    formatter = logging.Formatter(
        "%(asctime)s %(name)20s | %(levelname)s - %(message)s"
    )
    formatter.datefmt = '%H:%M:%S'
    return formatter


def init_logging(level: int | None = None):
    """ Set up a basic logger. """
    if level is None:
        level = logging.DEBUG if sys.flags.debug else logging.INFO

    logging.basicConfig(
        stream=sys.stdout,
        level=level,
        format=get_console_formatter()._fmt,
        datefmt=get_console_formatter().datefmt,
    )


""" 
Main Model
----------

This is the main model for the Segment-by-Segment application.
"""
from __future__ import annotations

import enum
import logging

from qtpy import QtCore
from qtpy.QtCore import Qt

from omc3_gui.segment_by_segment.measurement_model import OpticsMeasurement
from omc3_gui.segment_by_segment.segment_model import SegmentItemModel
from omc3_gui.ui_components.item_models import UniqueItemListModel

ItemDataRole = Qt.ItemDataRole
ItemFlag = Qt.ItemFlag

LOGGER = logging.getLogger(__name__)

class MeasurementListModel(QtCore.QAbstractListModel, UniqueItemListModel):

    _items: list[OpticsMeasurement]  # only for the IDE
    
    class ColorIDs(enum.IntEnum):
        NONE = 0
        BEAM1 = enum.auto()
        BEAM2 = enum.auto()
        RING1 = enum.auto()
        RING2 = enum.auto()
        RING3 = enum.auto()
        RING4 = enum.auto()

        @classmethod
        def get_color(cls, meas: OpticsMeasurement) -> int:
            if meas.accel == "lhc":
                return getattr(cls, f"BEAM{meas.beam}")
            
            if meas.accel == "psb":
                return getattr(cls, f"RING{meas.ring}")
            
            return cls.NONE

    def __init__(self, *args, **kwargs):
        super(QtCore.QAbstractListModel, self).__init__(*args, **kwargs)
        super(UniqueItemListModel, self).__init__()

    def data(self, index: QtCore.QModelIndex, role: int = ItemDataRole.DisplayRole):

        meas: OpticsMeasurement = self.get_item_at(index.row())
        # https://doc.qt.io/qt-5/qt.html#ItemDataRole-enum
        if role == ItemDataRole.DisplayRole:  
            return meas.display()

        if role == ItemDataRole.ToolTipRole:
            return meas.tooltip()

        if role == Qt.TextColorRole:
            return self.ColorIDs.get_color(meas)

        if role == ItemDataRole.UserRole:
            return meas

    def rowCount(self, index: QtCore.QModelIndex = None):
        return len(self._items)

    @property
    def items(self):
        return self._items


class SegmentTableModel(QtCore.QAbstractTableModel, UniqueItemListModel):
    """ Data Model for the table of segments. 
    
    Hint: Uses Qt.UserRole to retrieve the actual segment.
    """

    _COLUMNS: list[str] = ["Segment", "Start", "End"]  # display names
    _ATTRIBUTES: list[str] = ["name", "start", "end"]  # segment attributes
    
    _items: list[SegmentItemModel]  # only for the IDE
    
    def __init__(self, *args, **kwargs): 
        super(QtCore.QAbstractTableModel, self).__init__(*args, **kwargs)
        super(UniqueItemListModel, self).__init__()  # Items need to be unique

    def headerData(self, section, orientation, role=ItemDataRole.DisplayRole):
        """ Sets the header of the table. """
        # When we are displaying the header, use the display column names
        if orientation == QtCore.Qt.Horizontal and role == ItemDataRole.DisplayRole:
            return self._COLUMNS[section]

        # Otherwise whatever the default is    
        return super().headerData(section, orientation, role)

    def rowCount(self, parent=QtCore.QModelIndex()):
        """ Returns the number of rows in the model. """
        return len(self._items) 

    def columnCount(self, parent=QtCore.QModelIndex()):
        """ Returns the number of columns in the model. """
        return len(self._COLUMNS) 

    def data(self, index: QtCore.QModelIndex, role=ItemDataRole.DisplayRole):
        """ Return the data, depending on index and role. """
        i = index.row()
        j = index.column()
        segment: SegmentItemModel = self.get_item_at(i)
        
        if role == ItemDataRole.DisplayRole or role == ItemDataRole.EditRole:
            return str(getattr(segment, self._ATTRIBUTES[j]))
        
        if role == ItemDataRole.ToolTipRole:
            return segment.tooltip()

        if role == ItemDataRole.UserRole:
            return segment
        
    def setData(self, index, value, role):
        """ Set the data, depending on index and role. """
        i = index.row()
        j = index.column()
        segment: SegmentItemModel = self.get_item_at(i)

        if role == ItemDataRole.EditRole:
            if value is None or value == "":
                return False
            
            attribute = self._ATTRIBUTES[j]
            setattr(segment, attribute, value)

            self.dataChanged.emit(index, index)
            return True
        
    def flags(self, index):
        """ Set the flags for the given index. 
        At the moment: all elements are editable and selectable. """
        return ItemFlag.ItemIsEnabled | ItemFlag.ItemIsEditable | ItemFlag.ItemIsSelectable
